package ru.ifmo.genetics.dna;

import ru.ifmo.genetics.tools.Util;

/**
 * Immutable, 2-bits-per-nucleotide.
 */
public class Dna implements LightDna {
	private final byte[] value;
	private final int offset;
    private final int length;
	private final boolean rev;
	private final boolean compl;

    protected final static int sz_shift = 2;
    protected final static int sz = 1 << sz_shift;
    protected final static int sz_mod_mask = sz - 1;

	public Dna(byte[] value, int offset, int length, boolean rev, boolean compl) {
		this.value = value;
		this.offset = offset;
		this.length = length;
		this.rev = rev;
		this.compl = compl;
	}

    // one nuc per index
	public Dna(byte[] nucs) {
        this(nucs.length);
        for (int i = 0; i < nucs.length; ++i) {
            setNucInArray(this.value, i, nucs[i]);
        }
	}
	
	protected Dna(byte[] value, int offset, int length) {
		this(value, offset, length, false, false);
	}

	public Dna(int length) {
		this(newArray(length), 0, length);
	}

	public Dna(CharSequence s) {
		this(s.length());
		for (int i = 0; i < length; i++) {
			setNucInArray(this.value, i, DnaTools.fromChar(s.charAt(i)));
		}
	}

	/**
	 * Makes full DNA copy
	 */
	public Dna(LightDna dna) {
		this(dna.length());
		for (int i = 0; i < length; i++) {
			setNucInArray(this.value, i, dna.nucAt(i));
		}
	}

	/**
	 * Makes full copy of all DNAs
	 */
	public Dna(LightDna... dnas) {
		this(Util.sumOfDnaLengths(dnas));
		int x = 0;
		for (LightDna dna : dnas) {
			for (int i = 0; i < dna.length(); i++) {
				setNucInArray(this.value, x++, dna.nucAt(i));
			}
		}
	}

	public Dna reverse() {
		return new Dna(value, offset, length, !rev, compl);
	}

	public Dna complement() {
		return new Dna(value, offset, length, rev, !compl);
	}

	public Dna reverseComplement() {
		return new Dna(value, offset, length, !rev, !compl);
	}
	
	private static byte[] newArray(int length) {
		return new byte[(length + sz - 1) >> sz_shift];
	}

	/**
     * shouldn't be used when current nuc is not 0
	 * @param xxx index in the array, not in the DNA
	 */
	private static void setNucInArray(byte[] array, int xxx, byte value) {
		int i = xxx >> sz_shift;
		int j = xxx & sz_mod_mask;
        assert ((array[i] >>> (2 * j)) & sz_mod_mask) == 0;
		array[i] |= value << (2 * j);
	}

    @Override
    public int length() {
        return length;
    }

    public byte nucAt(int index) {
		int x = offset + (rev ? length - 1 - index: index);
		int i = x >> sz_shift;
		int j = x & sz_mod_mask;
		byte nuc = (byte) ((this.value[i] >>> (2 * j)) & 3);
		return (byte) (compl ? nuc ^ 3 : nuc);
	}
    
    public byte[] toByteArray() {
    	byte[] array = new byte[length()];
    	for (int i = 0; i < length; ++i) {
    		array[i] = nucAt(i);
    	}
    	return array;
    }

	public Dna substring(int beginIndex, int endIndex) {
		if (beginIndex == 0 && endIndex == length) {
			return this;
		}
		return new Dna(value, offset + (rev ? length - endIndex : beginIndex), endIndex - beginIndex, rev, compl);
	}
	
	@Override
	public String toString() {
	    StringBuilder sb = new StringBuilder();
	    for (int i = 0; i < length(); i++) {
	        sb.append(DnaTools.toChar(nucAt(i)));
	    }
	    return sb.toString();
	}

//	public Dna concat(Dna that) {
//		int length = count + that.count;
//		int[] value = newArray(length);
//		for (int i = 0; i < count; i++) {
//			setNucInArray(value, i, nucAt(i));
//		}
//		for (int i = 0; i < that.count; i++) {
//			setNucInArray(value, count + i, that.nucAt(i));
//		}
//		return new Dna(value, 0, length);
//	}
//
//	@Override
//	public int compareTo(IDna o) {
//		if (this == o) {
//			return 0;
//		}
//		int i = 0;
//		int j = 0;
//		// TODO faster?
//		while (i < count && j < o.length()) {
//			byte n1 = nucAt(i++);
//			byte n2 = o.nucAt(j++);
//			if (n1 != n2) {
//				return n1 - n2;
//			}
//		}
//		return count - o.length();
//	}


    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        Dna dna = (Dna) o;

        return DnaTools.equals(this, dna);
    }

    @Override
    public int hashCode() {
        return DnaTools.hashCode(this);
    }
}
