package ru.ifmo.genetics.dna;

import ru.ifmo.genetics.io.formats.Illumina;
import ru.ifmo.genetics.tools.Util;

public class DnaTools {
    private DnaTools() {
    }
    public final static char[] DNA = {'A', 'G', 'C', 'T'};
    public final static byte[] DNAcodes = {'A', 'G', 'C', 'T'};
    public final static byte[] sortedDNAcodes = {'A', 'C', 'G', 'T'};


    public static byte fromChar(char c) {
        switch (c) {
            case 'A':
                return 0;
            case 'G':
                return 1;
            case 'C':
                return 2;
            case 'T':
                return 3;
            default:
                throw new IllegalArgumentException("Incorrect nucleotide char: \"" + c + "\"");
        }
    }

    public static char toChar(byte b) {
        switch (b & 3) {
            case 0:
                return 'A';
            case 1:
                return 'G';
            case 2:
                return 'C';
            case 3:
                return 'T';
            default:
                throw new Error("World's gone crazy!");
        }
    }

    public static byte complement(byte b) {
        return (byte) (b ^ 3);
    }

    public static char complement(char c) {
        return toChar(complement(fromChar(c)));
    }


    /**
     * @param phred Quality score in Phred system
     * @return probability of nucleotide being correct
     */
    public static double probability(int phred) {
        return Math.max(1 - Math.pow(10, -phred * 0.1), 0.25);
    }

    public static String toString(LightDna dna) {
        char[] c = new char[dna.length()];
        for (int i = 0; i < dna.length(); i++) {
            c[i] = DnaTools.toChar(dna.nucAt(i));
        }
        return new String(c);
    }
    
    public static String toPhredString(LightDnaQ dnaq) {
        Illumina illumina = new Illumina();
        char[] c = new char[dnaq.length()];
        for (int i = 0; i < dnaq.length(); ++i) {
            c[i] = illumina.getPhredChar(dnaq.phredAt(i));
        }
        return new String(c);
    }
    
    public static boolean equals(LightDna dna1, LightDna dna2) {
        int m1 = dna1.length();
        int m2 = dna2.length();
        if (m1 != m2)
            return false;

        for (int i = 0; i < m1; ++i) {
            if (dna1.nucAt(i) != dna2.nucAt(i)) {
                return false;
            }
        }
        return true;
    }
    
    public static int hashCode(LightDna dna) {
        int length = dna.length();
        int res = 0;
        for (int i = 0; i < length; ++i) {
            res = 31 * res + dna.nucAt(i);
        }
        return res;
    }

    public static int compare(LightDna dna1, LightDna dna2) {
        int m = Math.min(dna1.length(), dna2.length());
        for (int i = 0; i < m; ++i) {
            int res = Util.compare(dna1.nucAt(i), dna2.nucAt(i));
            if (res != 0) {
                return res;
            }
        }

        return Util.compare(dna1.length(), dna2.length());
    }

    public static boolean contains(LightDna dna, String subsequence) {
        for (int i = 0; i <= dna.length() - subsequence.length(); ++i) {
            boolean ok = true;
            for (int j = 0; j < subsequence.length(); ++j) {
                if (toChar(dna.nucAt(i + j)) != subsequence.charAt(j)) {
                    ok = false;
                    break;
                }
            }
            if (ok) {
                return true;
            }
        }
        return false;
    }
}
