package ru.ifmo.genetics.tools;

import ru.ifmo.genetics.tools.executors.Latch;

import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.util.List;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.atomic.AtomicInteger;

public class DedicatedWriter implements Runnable {
    private PrintWriter out;
    
    private ConcurrentLinkedQueue<List<String>> queue;
    private AtomicInteger size;

    private Latch writingThreads;
    private volatile boolean finished;


    public DedicatedWriter(String fileName) throws FileNotFoundException {
        this(new PrintWriter(fileName));
    }
    public DedicatedWriter(PrintWriter out) {
        this.out = out;
        queue = new ConcurrentLinkedQueue<List<String>>();
        size = new AtomicInteger(0);
        writingThreads = new Latch();
        finished = false;
    }
    
    
    public void add(List<String> listToWrite) {
        assert !finished;

        queue.add(listToWrite);
        increaseAndNotify();
    }

    private void increaseAndNotify() {
        int prevSize = size.getAndIncrement();
        if (prevSize == 0) {
            synchronized (queue) {
                queue.notify();
            }
        }
    }

    /**
     * Call this method when no thread will add a list to write.
     * This call blocks until all list will be written.
     */
    public void close() {
        finished = true;
        increaseAndNotify();

        try {
            writingThreads.await();
        } catch (InterruptedException e) {
            System.err.println("Method await was interrupted!");
        }
        out.close();
    }

    @Override
    public void run() {
        writingThreads.increase();
        try {
            while (true) {
                List<String> list = queue.poll();
                while (list != null) {
                    for (String s : list) {
                        out.println(s);
                    }
                    size.decrementAndGet();
                    list = queue.poll();
                }

                if ((finished) && (queue.isEmpty())) {
                    break;
                }

                synchronized (queue) {
                    while ((queue.isEmpty()) && (!finished)) {
                        queue.wait();
                    }
                }
            }
        } catch (InterruptedException e) {
            System.err.println("Writing thread was interrupted!");
        } finally {
            writingThreads.decrease();
        }
    }
}
