package ru.ifmo.genetics.tools;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.configuration.Configuration;
import org.apache.commons.configuration.MapConfiguration;
import ru.ifmo.genetics.dna.DnaQ;
import ru.ifmo.genetics.dna.DnaTools;
import ru.ifmo.genetics.dna.LightDna;
import ru.ifmo.genetics.framework.Dataset;
import ru.ifmo.genetics.utils.pairs.UniPair;
import ru.ifmo.genetics.io.formats.Illumina;

import java.io.*;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.TimeUnit;

public class Util {

    public static void datasets2BinqFiles(Iterable<Dataset> datasets, String dir)
            throws IOException {
        for (Dataset dataset : datasets) {
            dataset2BinqFile(dataset, dir);
        }
    }

    public static void datasets2FastqFiles(Iterable<Dataset> datasets, String dir)
            throws IOException {
        for (Dataset dataset : datasets) {
            dataset2FastqFile(dataset, dir);
        }
    }

    public static void putInt(int v, OutputStream out) throws IOException {
        out.write((v >>> 24) & 0xFF);
        out.write((v >>> 16) & 0xFF);
        out.write((v >>> 8) & 0xFF);
        out.write((v >>> 0) & 0xFF);
    }

    public static void putByteArray(byte[] array, OutputStream out) throws IOException {
        putInt(array.length, out);
        out.write(array);
    }

    public static int getInt(InputStream in) throws IOException {
        int ch1 = in.read();
        int ch2 = in.read();
        int ch3 = in.read();
        int ch4 = in.read();
        if ((ch1 | ch2 | ch3 | ch4) < 0)
            throw new EOFException();
        return (ch1 << 24) + (ch2 << 16) + (ch3 << 8) + (ch4 << 0);
    }

    public static byte[] getByteArray(InputStream in) throws IOException {
        int len = getInt(in);
        byte[] b = new byte[len];
        in.read(b);
        return b;
    }

    public static long getPrefixCode(String prefix, int len) {
        long res = 0;
        for (int i = 0; i < prefix.length(); i++) {
            res = (res << 2) | DnaTools.fromChar(prefix.charAt(i));
        }
        res = res << (2 * (len - prefix.length()));
        return res;
    }

    public static long getPrefixMask(String prefix, int len) {
        long res = 0;
        for (int i = 0; i < prefix.length(); i++) {
            res = (res << 2) | 3;
        }
        res = res << (2 * (len - prefix.length()));
        return res;
    }

    public static long getCode(String string) {
        long res = 0;
        for (int i = 0; i < string.length(); i++) {
            res = (res << 2) | DnaTools.fromChar(string.charAt(i));
        }
        return res;
    }

    public static void dnaQs2BinqFile(Iterable<DnaQ> dnaqs, String datasetName,
                                      String filename) throws IOException {
        dnaQs2BinqFile(dnaqs, datasetName, filename, false);
    }

    public static void dnaQs2BinqFile(Iterator<DnaQ> dnaqsIt, String datasetName,
                                      String filename, boolean append) throws IOException {
        OutputStream out = new BufferedOutputStream(new FileOutputStream(filename, append));
        while (dnaqsIt.hasNext()) {
            putByteArray(dnaqsIt.next().toByteArray(), out);
        }
        out.close();
    }

    public static void dnaQs2BinqFile(Iterable<DnaQ> dnaqs, String datasetName,
                                      String filename, boolean append) throws IOException {
        dnaQs2BinqFile(dnaqs.iterator(), datasetName, filename, append);
    }

    public static void dataset2BinqFile(Dataset dataset, String dir) throws IOException {
        String fullDatasetName = dir + File.separator + dataset.name();
        System.err.println("fullDatasetName = " + fullDatasetName);

        System.err.print(fullDatasetName + "_1.binq...");
        dnaQs2BinqFile(dataset.allFirsts(), fullDatasetName, fullDatasetName + "_1.binq");
        System.err.println(" done");

        System.err.print(fullDatasetName + "_2.binq...");
        dnaQs2BinqFile(dataset.allSeconds(), fullDatasetName, fullDatasetName + "_2.binq");
        System.err.println(" done");
    }

    public static void dnaQs2FastqFile(Iterable<DnaQ> dnaqs, String datasetName,
                                       String filename) throws IOException {
        dnaQs2FastqFile(dnaqs, datasetName, filename, false, 0);
    }

    public static void dnaQs2FastqFile(Iterable<DnaQ> dnaqs, String datasetName,
                                       String filename, boolean append, long startNumber) throws IOException {
        Illumina il = new Illumina();
        PrintWriter out = new PrintWriter(new FileOutputStream(filename, append));
        long i = startNumber;
        for (DnaQ dnaq : dnaqs) {
            out.println("@" + datasetName + ":" + ++i + "#0/1");
            out.println(Util.DnaQ2String(dnaq));
            out.println("+" + datasetName + ":" + i + "#0/1");
            for (int j = 0; j < dnaq.length(); ++j) {
                out.print(il.getPhredChar(dnaq.phredAt(j)));
            }
            out.println();
        }
        out.close();
    }

    public static void dnaQs2DoubleFastaFile(Iterable<DnaQ> dnaqs, String filePrefix)
            throws IOException {
        dnaQs2DoubleFastaFile(dnaqs, filePrefix, false, 0);
    }

    public static void dnaQs2DoubleFastaFile(Iterable<DnaQ> dnaqs,
                                             String filePrefix, boolean append, long startNumber)
            throws IOException {
        Illumina il = new Illumina();
        PrintWriter out1 = new PrintWriter(new FileOutputStream(filePrefix + ".fasta", append));
        PrintWriter out2 = new PrintWriter(new FileOutputStream(filePrefix + ".qual", append));
        long i = startNumber;
        String baseName = filePrefix;
        int id = 1;
        if (filePrefix.endsWith("_1") || filePrefix.endsWith("_2")) {
            baseName = baseName.substring(0, baseName.length() - 2);
            id = filePrefix.endsWith("_2") ? 2 : 1;
        }
        for (DnaQ dnaq : dnaqs) {
            out1.println(">" + baseName + ":" + i + "#0/" + id);
            out2.println(">" + baseName + ":" + i + "#0/" + id);
            ++i;
            out1.println(Util.DnaQ2String(dnaq));
            for (int j = 0; j < dnaq.length(); ++j) {
                out2.print(dnaq.phredAt(j) + " ");
            }
            out2.println();
        }
        out1.close();
        out2.close();
    }

    public static void dataset2FastqFile(Dataset dataset, String dir) throws IOException {
        String fullDatasetName = dir + File.separator + dataset.name();

        System.err.print(fullDatasetName + "_1.fastq...");
        dnaQs2FastqFile(dataset.allFirsts(), fullDatasetName, fullDatasetName + "_1.fastq");
        System.err.println(" done");

        System.err.print(fullDatasetName + "_2.fastq...");
        dnaQs2FastqFile(dataset.allSeconds(), fullDatasetName, fullDatasetName + "_2.fastq");
        System.err.println(" done");
    }

    public static <K> void addInt(Map<K, Integer> map, K key, int value) {
        Integer t = map.get(key);
        if (t == null)
            t = 0;
        map.put(key, t + value);
    }

    public static <K> void addLong(Map<K, Long> map, K key, long value) {
        Long t = map.get(key);
        if (t == null)
            t = 0L;
        map.put(key, t + value);
    }

    public static <K> void incrementInt(Map<K, Integer> map, K key) {
        addInt(map, key, 1);
    }

    public static <K> void incrementLong(Map<K, Long> map, K key) {
        addLong(map, key, 1L);
    }

    public static long DnaQ2Long(DnaQ d) {
        int l = d.length();
        long r = 0;
        for (int i = 0; i < l; ++i) {
            r = (r << 2) | (d.byteAt(i) & 3);
        }
        return r;
    }

    public static String DnaQ2String(DnaQ d) {
        int l = d.length();
        StringBuilder s = new StringBuilder();
        for (int i = 0; i < l; ++i) {
            s.append(DnaTools.toChar(d.byteAt(i)));
        }
        return s.toString();
    }

    public static String join(Iterable<String> ss, String delimiter) {
        StringBuilder sb = new StringBuilder();
        boolean firstly = true;
        for (String s : ss) {
            if (!firstly) {
                sb.append(delimiter);
            }
            sb.append(s);
            firstly = false;
        }
        return sb.toString();
    }

    public static void shutdownAndAwaitTermination(ExecutorService pool) {
        pool.shutdown(); // Disable new tasks from being submitted
        try {
            // Wait a while for existing tasks to terminate
            while (!pool.awaitTermination(60, TimeUnit.SECONDS)) ;
        } catch (InterruptedException ie) {
            // (Re-)Cancel if current thread also interrupted
            pool.shutdownNow();
            // Preserve interrupt status
            Thread.currentThread().interrupt();
        }
    }

    public static int sumOfDnaLengths(LightDna... dnas) {
        int sum = 0;
        for (LightDna dna : dnas) {
            sum += dna.length();
        }
        return sum;
    }

    public static <E> Iterable<E> extractFirsts(final Iterable<UniPair<E>> pairs) {
        return new Iterable<E>() {
            public Iterator<E> iterator() {
                return new Iterator<E>() {
                    Iterator<UniPair<E>> pairIt = pairs.iterator();

                    public boolean hasNext() {
                        return pairIt.hasNext();
                    }

                    public E next() {
                        return pairIt.next().first;
                    }

                    public void remove() {
                        pairIt.remove();
                    }
                };
            }
        };
    }

    public static <E> Iterable<E> extractSeconds(final Iterable<UniPair<E>> pairs) {
        return new Iterable<E>() {
            public Iterator<E> iterator() {
                return new Iterator<E>() {
                    Iterator<UniPair<E>> pairIt = pairs.iterator();

                    public boolean hasNext() {
                        return pairIt.hasNext();
                    }

                    public E next() {
                        return pairIt.next().second;
                    }

                    public void remove() {
                        pairIt.remove();
                    }
                };
            }
        };
    }

    public static void swap(int[] x, int i, int j) {
        int t = x[i];
        x[i] = x[j];
        x[j] = t;
    }

    public static void swap(short[] x, int i, int j) {
        short t = x[i];
        x[i] = x[j];
        x[j] = t;
    }


    // replaces "-" with "_"
    public static void addOptionToConfig(CommandLine cmd, Configuration config, String option) {
        addOptionToConfig(cmd, config, option, option.replace("-", "_"));
    }

    public static void addOptionToConfig(
            CommandLine cmd,
            Configuration config,
            String option,
            String property) {
        if (cmd.hasOption(option)) {
            config.setProperty(property, cmd.getOptionValue(option));
        }
    }

    public static Configuration mergeConfigurations(Configuration config, Configuration globalConfig) {
        Configuration result = new MapConfiguration(new HashMap());

        addAllWithReplacement(result, globalConfig);
        addAllWithReplacement(result, config);

        return result;
    }
    
    public static void addAllWithReplacement(Configuration toConfig, Configuration fromConfig) {
        if (fromConfig == null) {
            return;
        }

        @SuppressWarnings("unchecked")
        Iterator<String> iterator = fromConfig.getKeys();
        while (iterator.hasNext()) {
            String key = iterator.next();
            toConfig.addProperty(key, fromConfig.getProperty(key));
        }
    }


    public static int compare(long a, long b) {
        return (a < b) ? -1 : (a == b ? 0 : 1);
    }

    public static int compare(int a, int b) {
        return (a < b) ? -1 : (a == b ? 0 : 1);
    }

    public static int compare(byte a, byte b) {
        return a - b;
    }

    public static int compare(boolean a, boolean b) {
        return (a == b) ? 0 : (a ? 1 : -1);
    }
}

