package ru.ifmo.genetics.tools.assembling.task;

import org.apache.hadoop.mapred.Reporter;
import ru.ifmo.genetics.dna.*;
import ru.ifmo.genetics.utils.pairs.UniPair;
import ru.ifmo.genetics.io.formats.Illumina;
import ru.ifmo.genetics.statistics.Timer;
import ru.ifmo.genetics.tools.CompactDeBruijnGraph;
import ru.ifmo.genetics.utils.KmerUtils;

import java.io.PrintWriter;
import java.util.*;

public class FillingTask implements Runnable {
    private static byte DEFINETELY_QUALITY = 62;
    private static byte POLYMORPHISM_QUALITY = 2;
    private static int MAX_POLYMORPHISM_NUMBER = 20;

    private static int MAX_LAYER_SIZE = (int) 1e3;
    private final static boolean EDGE_SKIPPING_ENABLED = false;

    private static Illumina illumina = new Illumina();

    private static final double PHRED_THRESHOLD = 0.90;
    private static final double RIGHTNESS_METRIC_THRESHOLD = 1e-6;

    GlobalContext context;
    List<UniPair<DnaQ>> task;

    int ok = 0;
    int notFound = 0;
    int noAnchor = 0;
    int ambiguous = 0;
    int tooBig = 0;
    int tooShort = 0;
    int dropped = 0;
    int tooPolymorphic = 0;
    int processed = 0;

    public int getAmbiguous() {
        return ambiguous;
    }

    public void setAmbiguous(int ambiguous) {
        this.ambiguous = ambiguous;
    }

    public int getOk() {
        return ok;
    }

    public int getProcessed() {
        return processed;
    }

    public int getNotFound() {
        return notFound;
    }

    public enum ResultType { OK, NOT_FOUND, NO_ANCHOR, AMBIGUOUS, TOO_BIG, TOO_SHORT, TOO_POLYMORPHIC, FAIL}

    PrintWriter graphOut = null;


    public static class FillingResult {
        public DnaQ dnaq;
        public int leftSkip;
        public int rightSkip;
        public ResultType type;

        public FillingResult(ResultType type) {
            this.type = type;
        }

        public FillingResult(DnaQ dnaq, int leftSkip, int rightSkip) {
            this.dnaq = dnaq;
            this.leftSkip = leftSkip;
            this.rightSkip = rightSkip;
            this.type = ResultType.OK;
        }
    }

    public FillingTask(GlobalContext context, List<UniPair<DnaQ>> task) {
        this.context = context;
        this.task = task;
    }

    @Override
    public void run() {
        Timer t = new Timer();
        t.start();

        List<DnaQ> result = new ArrayList<DnaQ>(task.size());
        List<UniPair<DnaQ>> failed = new ArrayList<UniPair<DnaQ>>(task.size());


        processed = 0;
        for (UniPair<DnaQ> p : task) {
            //System.err.println(p.first);
            FillingResult res = fillRead(p.first, p.second, true);
            DnaQ dna = null;
            if (res != null) {
                dna = res.dnaq;
            }
            int oldNoAnchor = noAnchor;
            if (dna != null) {
                result.add(dna);
            }
            // else if (noAnchor == 1) {
            // System.err.println("First no anchor");
            // failed.add(p);
            // System.exit(0);
            // }

            // System.err.print("\r" + dna);
            // System.err.println(i + ": ok = " + ok + ", notfound = " + notFound + 
            // ", ambiguous = " + ambiguous + ", tooBig = " + tooBig + ", task.size = " + task.size());
        }

        System.err.println("took " + t.finish());
        printStat();

        context.queue.add(result);
        context.failedQueue.add(failed);
    }

    public void printStat() {
        System.err.println("ok = " + ok + ", notFound = " + notFound +
                ", noAnchor = " + noAnchor + ", ambiguous = " + ambiguous +
                ", tooBig = " + tooBig + ", tooPolymorphic = " + tooPolymorphic +
                ", tooShort = " + tooShort + ", processed = " + processed +
                ", dropped = " + dropped);
    }

    public void updateCounters(Reporter reporter) {
        reporter.getCounter("FillingTask", "ok").increment(ok);
        reporter.getCounter("FillingTask", "notFound").increment(notFound);
        reporter.getCounter("FillingTask", "noAnchor").increment(noAnchor);
        reporter.getCounter("FillingTask", "ambiguous").increment(ambiguous);
        reporter.getCounter("FillingTask", "tooBig").increment(tooBig);
        reporter.getCounter("FillingTask", "tooPolymorphic").increment(tooPolymorphic);
        reporter.getCounter("FillingTask", "tooShort").increment(tooShort);
        reporter.getCounter("FillingTask", "processed").increment(processed);
        reporter.getCounter("FillingTask", "dropped").increment(dropped);
    }


    public FillingResult fillRead(IDnaQ left, IDnaQ right) {
        return fillRead(left, right, true);
    }

    public FillingResult fillRead(LightDnaQ left, LightDnaQ right, boolean pairEnds) {
        ++processed;

        // System.err.println("from left:");
        // System.err.println(DnaTools.toString(left));
        // System.err.println("from right:");
        // System.err.println(DnaTools.toString(right));
        // System.err.println("to right:");
        // System.err.println(DnaTools.toString(new DnaView(right, 0, right.length(), true, true)));
//
        if ((left.length() < context.k) || (right.length() < context.k)) {
            tooShort++;
            return new FillingResult(ResultType.TOO_SHORT);
        }

        if (!pairEnds) {
            // mate pairs
            left = DnaQView.rcView(left);
            right = DnaQView.rcView(right);
        }

        int oldLength = right.length();
        right = DnaQView.complementView(right);
        int newLength = right.length();
        assert oldLength == newLength : oldLength + " " + newLength;

        long leftVertex = KmerUtils.toLong(left, 0, context.k);
        long rightVertex = KmerUtils.toLong(new DnaQView(right, 0, context.k, true, false));

        DnaQView leftHint = new DnaQView(left, context.k, left.length());
        DnaQView rightHint = new DnaQView(right, context.k, right.length());

        FillingResult res = doubleBfs(leftVertex, rightVertex,
                context.minFragmentSize - context.k,
                context.maxFragmentSize - context.k,
                leftHint, rightHint);

        if (res.type == ResultType.OK) {
            int polymorphismNumber = 0;
            for (int i = 0; i < res.dnaq.length(); ++i) {
                if (res.dnaq.phredAt(i) == POLYMORPHISM_QUALITY) {
                    ++polymorphismNumber;
                }
            }
            if (polymorphismNumber > MAX_POLYMORPHISM_NUMBER) {
                tooPolymorphic++;
                return new FillingResult(ResultType.TOO_POLYMORPHIC);
            }
            ok++;
        }

        return res;
    }

    private boolean buildReversePath(DnaQBuilder builder, Set<BfsTreeNode> layer) {
        boolean[] possibleBases = new boolean[4];
        for (BfsTreeNode n : layer) {
            possibleBases[(int) (n.v & 3)] = true;
        }

        int basesCount = 0;
        byte base = -1;

        for (byte i = 0; i < 4; ++i) {
            if (possibleBases[i]) {
                ++basesCount;
                if (base == -1) {
                    base = i;
                }
            }
        }

        /*
        if (basesCount > 2) {
            ++ambiguous;
            return false;
        }
        */

        Set<BfsTreeNode> nextLayer = new HashSet<BfsTreeNode>(layer.size());

        for (BfsTreeNode n : layer) {
            for (int i = 0; i < 4; ++i) {
                if (n.prev[i] != null) {
                    nextLayer.add(n.prev[i]);
                    if (graphOut != null) {
                        graphOut.printf(
                                "%s -> %s [ label = \"%s\"];\n",
                                dotNodeId(n.prev[i]),
                                dotNodeId(n),
                                DnaTools.toChar((byte) (n.v & 3)));
                    }
                }
            }
        }

        if (nextLayer.isEmpty()) {
            long vertex = layer.iterator().next().v;
            for (int j = context.k - 1; j >= 0; --j) {
                builder.append((byte) ((vertex >>> (2 * j)) & 3), DEFINETELY_QUALITY);
            }

            return true;
        }

        if (!buildReversePath(builder, nextLayer)) {
            return false;
        }

        builder.append(base, (basesCount == 1) ? DEFINETELY_QUALITY : POLYMORPHISM_QUALITY);
        return true;
    }

    private boolean buildPath(DnaQBuilder builder, Set<BfsTreeNode> layer) {
        Set<BfsTreeNode> currentLayer = new HashSet<BfsTreeNode>();
        Set<BfsTreeNode> nextLayer = new HashSet<BfsTreeNode>();

        currentLayer.addAll(layer);

        boolean[] possibleBases = new boolean[4];

        while (true) {
            Arrays.fill(possibleBases, false);

            for (BfsTreeNode n : currentLayer) {
                for (int i = 0; i < 4; ++i) {
                    possibleBases[i] |= (n.prev[i] != null);
                    if (graphOut != null && n.prev[i] != null) {
                        graphOut.println(dotNodeId(n) + " -> " + dotNodeId(n.prev[i]));
                        graphOut.printf(
                                "%s -> %s [ label = \"%s\"];\n",
                                dotNodeId(n),
                                dotNodeId(n.prev[i]),
                                DnaTools.toChar((byte) (n.prev[i].v & 3)));
                    }
                }
            }

            int basesCount = 0;
            byte base = -1;

            for (byte i = 0; i < 4; ++i) {
                if (possibleBases[i]) {
                    ++basesCount;
                    if (base == -1) {
                        base = i;
                    }
                }
            }

            /*
            if (basesCount > 2) {
                ++ambiguous;
                return false;
            }
            */

            if (basesCount == 0) {
                break;
            }

            builder.append(base, (basesCount == 1) ? DEFINETELY_QUALITY : POLYMORPHISM_QUALITY);

            for (BfsTreeNode n : currentLayer) {
                for (int i = 0; i < 4; ++i) {
                    if (n.prev[i] != null) {
                        nextLayer.add(n.prev[i]);
                    }
                }
            }

            Set<BfsTreeNode> temp = currentLayer;
            currentLayer = nextLayer;
            nextLayer = temp;
            nextLayer.clear();
        }
        return true;
    }


    private void leftBfsTurn(Map<Long, BfsTreeNode> currentQueue,
                             Map<Long, BfsTreeNode> nextQueue, byte nuc, double rightnessFactor) {
        boolean useHint = (rightnessFactor > PHRED_THRESHOLD);
        for (Map.Entry<Long, BfsTreeNode> entry : currentQueue.entrySet()) {
            long v = entry.getKey();
            long e = v << 2;
            for (int i = 0; i < 4; ++i, ++e) {
                double localRightnessFactor = 1;
                if (useHint) {
                    if (i == nuc) {
                        localRightnessFactor = rightnessFactor;
                    } else {
                        localRightnessFactor = (1 - rightnessFactor) / 3;
                    }
                }
                boolean shouldBeAdded = context.graph.containsEdge(e);
                if (EDGE_SKIPPING_ENABLED && !shouldBeAdded) {
                    long e2 = (e & context.toVertexMask) << 2;
                    for (int j2 = 0; j2 < 4; ++j2, ++e2) {
                        if (context.graph.containsEdge(e2)) {
                            shouldBeAdded = true;
                            break;
                        }
                    }
                }

                if (shouldBeAdded) {
                    long newVertex = e & context.toVertexMask;

                    BfsTreeNode node = nextQueue.get(newVertex);
                    if (node == null) {
                        node = new BfsTreeNode(newVertex);
                        nextQueue.put(newVertex, node);
                    }
                    node.prev[(int) (e >>> (2 * context.k))] = entry.getValue();
                    node.rightnessMetric += entry.getValue().rightnessMetric * localRightnessFactor;
                }
            }
        }
    }

    private void rightBfsTurn(Map<Long, BfsTreeNode> currentQueue,
                              Map<Long, BfsTreeNode> nextQueue, byte nuc, double rightnessFactor) {
        boolean useHint = (rightnessFactor > PHRED_THRESHOLD);
        for (Map.Entry<Long, BfsTreeNode> entry : currentQueue.entrySet()) {
            long v = entry.getKey();
            long e = v;
            for (int i = 0; i < 4; ++i, e += context.toVertexMask + 1) {
                double localRightnessFactor = 1;
                if (useHint) {
                    if (i == nuc) {
                        localRightnessFactor = rightnessFactor;
                    } else {
                        localRightnessFactor = (1 - rightnessFactor) / 3;
                    }
                }
                boolean shouldBeAdded = context.graph.containsEdge(e);
                if (EDGE_SKIPPING_ENABLED && !shouldBeAdded) {
                    long e2 = e >>> 2;
                    for (int j2 = 0; j2 < 4; ++j2, e2 += context.toVertexMask + 1) {
                        if (context.graph.containsEdge(e2)) {
                            shouldBeAdded = true;
                            break;
                        }
                    }
                }

                if (shouldBeAdded) {
                    long newVertex = e >>> 2;

                    BfsTreeNode node = nextQueue.get(newVertex);
                    if (node == null) {
                        node = new BfsTreeNode(newVertex);
                        nextQueue.put(newVertex, node);
                    }
                    node.prev[(int) (v & 3)] = entry.getValue();
                    node.rightnessMetric += entry.getValue().rightnessMetric * localRightnessFactor;
                }
            }
        }
    }

    private void removeIncorrectNodes(Map<Long, BfsTreeNode> queue) {
        Iterator<BfsTreeNode> it = queue.values().iterator();
        while (it.hasNext()) {
            BfsTreeNode node = it.next();
            if (node.rightnessMetric < RIGHTNESS_METRIC_THRESHOLD) {
                dropped++;
                it.remove();
            }
        }
    }

    private FillingResult doubleBfs(long leftVertex,
                                    long rightVertex, long minPathLength, long maxPathLength,
                                    LightDnaQ leftHint, LightDnaQ rightHint) {
        DnaQBuilder leftDnaBuilder = new DnaQBuilder(context.dnaBuilderCapacity);
        Map<Long, BfsTreeNode> leftQueue = new TreeMap<Long, BfsTreeNode>();
        Map<Long, BfsTreeNode> leftNextQueue = new TreeMap<Long, BfsTreeNode>();
        Map<Long, BfsTreeNode> rightQueue = new TreeMap<Long, BfsTreeNode>();
        Map<Long, BfsTreeNode> rightNextQueue = new TreeMap<Long, BfsTreeNode>();

        leftQueue.put(leftVertex, new BfsTreeNode(leftVertex, 1));
        rightQueue.put(rightVertex, new BfsTreeNode(rightVertex, 1));

        /*
        try {
            graphOut = new PrintWriter("graph" + hashCode() + "_" + processed);
        } catch (FileNotFoundException e) {
            e.printStackTrace();
        }
        */

        if (graphOut != null) {
            graphOut.println("digraph G" + hashCode() + "_" + processed + "{");
            graphOut.println("center=true; rankdir=LR;");
        }

        DnaQ res = null;
        int leftSkipped = -1;
        int rightSkipped = -1;

        int leftLength = 0;
        int rightLength = 0;

        int maxSize = 2;

        boolean isAmbiguous = false;

        // printLeftLayer(leftQueue.values(), 0);
        // printRightLayer(rightQueue.values(), (int) (maxPathLength + 1));
        boolean noLeftAnchor = true;
        boolean noRightAnchor = true;


        for (int length = 1; length <= maxPathLength; ++length) {
            /*

            System.err.println("iteration: " + length);
            for (Long l : leftQueue.keySet()) {
                System.err.print("L");
                for (int i = 0; i < leftLength; ++i)
                    System.err.print(' ');
                System.err.println(KmerUtils.kmer2String(l, context.k));
            }
            for (Long l : rightQueue.keySet()) {
                System.err.print("R");
                for (int i = 0; i < maxPathLength - rightLength; ++i)
                    System.err.print(' ');
                System.err.println(KmerUtils.kmer2String(l, context.k));
            }
            */


            byte nuc = 0;
            byte phred = 2;
            if (leftQueue.size() < rightQueue.size() || (leftQueue.size() == rightQueue.size() && (length & 1) == 0)) { // left
                // System.err.println("left");
                if (leftLength < leftHint.length()) {
                    nuc = leftHint.nucAt(leftLength);
                    phred = leftHint.phredAt(leftLength);
                }
                leftBfsTurn(leftQueue, leftNextQueue, nuc, 1 - illumina.getProbability(phred));

                Map<Long, BfsTreeNode> oldQueue = leftQueue;
                leftQueue = leftNextQueue;
                leftNextQueue = oldQueue;

                leftNextQueue.clear();


                if (!leftQueue.isEmpty()) {
                    noLeftAnchor = false;
                }

                if (leftLength < leftHint.length()) {
                    leftVertex = ((leftVertex << 2) + nuc) & context.toVertexMask;
                    BfsTreeNode oldNode = leftQueue.get(leftVertex);
                    if (oldNode != null) {
                        oldNode.rightnessMetric = 1;
                    } else {
                        leftQueue.put(leftVertex, new BfsTreeNode(leftVertex, 1));
                    }
                }

                leftLength++;

                if (leftQueue.isEmpty()) {
                    break;
                }

                // printLeftLayer(leftQueue.values(), leftLength);

                if (leftLength < leftHint.length()) {
                    removeIncorrectNodes(leftQueue);
                }

            } else { // right
                // System.err.println("right");
                if (rightLength < rightHint.length()) {
                        nuc = rightHint.nucAt(rightLength);
                        phred = rightHint.phredAt(rightLength);
                }
                rightBfsTurn(rightQueue, rightNextQueue, nuc, 1 - illumina.getProbability(phred));

                Map<Long, BfsTreeNode> oldQueue = rightQueue;
                rightQueue = rightNextQueue;
                rightNextQueue = oldQueue;

                rightNextQueue.clear();


                if (!rightQueue.isEmpty()) {
                    noRightAnchor = false;
                }

                if (rightLength < rightHint.length()) {
                    rightVertex = (rightVertex + ((long) nuc << (2 * context.k))) >>> 2;
                    BfsTreeNode oldNode = rightQueue.get(rightVertex);
                    if (oldNode != null) {
                        oldNode.rightnessMetric = 1;
                    } else {
                        rightQueue.put(rightVertex, new BfsTreeNode(rightVertex, 1));
                    }
                }

                rightLength++;

                if (rightQueue.isEmpty()) {
//                    System.err.println("empty");
                    break;
                }
                // printRightLayer(rightQueue.values(), (int)(maxPathLength + 1 - rightLength));
                if (rightLength < rightHint.length()) {
                    removeIncorrectNodes(rightQueue);
                }
            }

            // System.err.println(length + ": " + leftQueue.size() + " + " + rightQueue.size());

            if (length >= minPathLength) {
                Set<BfsTreeNode> leftHalfOfIntersection = new HashSet<BfsTreeNode>();
                Set<BfsTreeNode> rightHalfOfIntersection = new HashSet<BfsTreeNode>();
                for (Map.Entry<Long, BfsTreeNode> e : leftQueue.entrySet()) {
                    if (rightQueue.containsKey(e.getKey())) {
                        leftHalfOfIntersection.add(e.getValue());
                        rightHalfOfIntersection.add(rightQueue.get(e.getKey()));
                        /*
                        if (graphOut != null) {
                            graphOut.println(
                                    dotNodeId(e.getValue(), leftLength) + " -> " +
                                    dotNodeId(rightQueue.get(e.getKey()), (int)(maxPathLength + 1 - rightLength)) +
                                    "[ color = \"green\", arrowhead=none ];");
                        }
                        */
                    }
                }

                if (!leftHalfOfIntersection.isEmpty()) {
                    if (res != null) {
                        // isAmbiguous = true;
                        ambiguous++;
//                        System.err.println("ambiguous");
                        return new FillingResult(ResultType.AMBIGUOUS);
                    }

                    if (!buildReversePath(leftDnaBuilder, leftHalfOfIntersection)) {
//                        System.err.println("fail1");
                        return new FillingResult(ResultType.FAIL);
                    }

                    leftSkipped = leftLength + context.k - leftDnaBuilder.length();

                    int oldLength = leftDnaBuilder.length();

                    if (!buildPath(leftDnaBuilder, rightHalfOfIntersection)) {
//                        System.err.println("fail2");
                        return new FillingResult(ResultType.FAIL);
                    }

                    rightSkipped = rightLength - (leftDnaBuilder.length() - oldLength);

                    res = leftDnaBuilder.build();
                    // System.err.println("built:");
                    // System.err.println(res);
                    leftDnaBuilder = new DnaQBuilder(context.dnaBuilderCapacity);
                }
            }

            maxSize = Math.max(maxSize, leftQueue.size() + rightQueue.size());

            if (maxSize > MAX_LAYER_SIZE) {
                if (graphOut != null) {
                    graphOut.println("}");
                    graphOut.close();
                }
                tooBig++;
//                System.err.println("toobig");
                return new FillingResult(ResultType.TOO_BIG);
            }


        }

        if (graphOut != null) {
            graphOut.println("}");
            graphOut.close();
        }

        if (isAmbiguous) {

            ambiguous++;
            return new FillingResult(ResultType.AMBIGUOUS);
        }

        if (leftLength > 0 && noLeftAnchor || rightLength > 0 && noRightAnchor) {
            noAnchor++;
            return new FillingResult(ResultType.NO_ANCHOR);
        }

        if (res == null) {
            notFound++;
            return new FillingResult(ResultType.NOT_FOUND);
        }

        return new FillingResult(res, leftSkipped, rightSkipped);
    }

    private static class BfsTreeNode {
        public long v;
        public BfsTreeNode[] prev = new BfsTreeNode[4];
        public double rightnessMetric;

        private BfsTreeNode(long v, double rightnessMetric) {
            this.v = v;
            this.rightnessMetric = rightnessMetric;
        }

        public BfsTreeNode(long v) {
            this(v, 0);
        }

        public String toString(int len) {
            return KmerUtils.kmer2String(v, len);
        }
    }

    private String dotNodeId(BfsTreeNode node, int level) {
        return "node" + node.toString(context.k) + "_" + level;
    }

    private String dotNodeId(BfsTreeNode node) {
        return node.toString(context.k);
    }

    private void printLayer(Collection<BfsTreeNode> layer, int i) {
        for (BfsTreeNode node : layer) {
            graphOut.printf(
                    "%s [ label = \"%s\", shape=point];\n",
                    dotNodeId(node, i),
                    node.toString(context.k)
            );
        }

        graphOut.print("{ rank=same; ");
        for (BfsTreeNode node : layer) {
            graphOut.print(dotNodeId(node, i));
            graphOut.print("; ");
        }
        graphOut.println("}");

    }

    private void printLeftLayer(Collection<BfsTreeNode> layer, int i) {
        if (graphOut == null)
            return;
        printLayer(layer, i);
        for (BfsTreeNode node : layer) {
            for (int j = 0; j < 4; ++j) {
                if (node.prev[j] != null) {
                    graphOut.printf(
                            "%s -> %s [ label = \"%s\", color=blue];\n",
                            dotNodeId(node.prev[j], i - 1),
                            dotNodeId(node, i),
                            DnaTools.toChar((byte) (node.v & 3)));
                }
            }
        }
    }

    private void printRightLayer(Collection<BfsTreeNode> layer, int i) {
        if (graphOut == null)
            return;
        printLayer(layer, i);
        for (BfsTreeNode node : layer) {
            for (int j = 0; j < 4; ++j) {
                if (node.prev[j] != null) {
                    graphOut.printf(
                            "%s -> %s [ label = \"%s\", color=red];\n",
                            dotNodeId(node, i),
                            dotNodeId(node.prev[j], i + 1),
                            DnaTools.toChar((byte) (node.v & 3)));
                }
            }
        }
    }

    public static void main(String[] args) {
        // 3 3 0 1 1 0 2|0 3 0 2 0 0
        // T T A G G A C A T A C A A
        int k = 2;
        CompactDeBruijnGraph graph = new CompactDeBruijnGraph(k, 128 * 8);
        graph.addEdge(3 * 16L + 0 * 4 + 1);
        graph.addEdge(0 * 16L + 1 * 4 + 1);
        graph.addEdge(1 * 16L + 1 * 4 + 0);
        graph.addEdge(1 * 16L + 0 * 4 + 2);
        graph.addEdge(0 * 16L + 2 * 4 + 3);
        graph.addEdge(2 * 16L + 3 * 4 + 0);
        graph.addEdge(3 * 16L + 0 * 4 + 2);
        graph.addEdge(0 * 16L + 2 * 4 + 0);

        graph.addEdge(1 * 16L + 0 * 4 + 0);
        graph.addEdge(0 * 16L + 0 * 4 + 3);
        graph.addEdge(0 * 16L + 3 * 4 + 0);


        /*
        AAT
        ACA
        ACT
        AGG
        AGT
        ATA
        ATT
        CCT
        CTA
        CTA
        GAA
        GAC
        GGA
        GTA
        GTC
        TAC
        TAG
        TAG
        TAT
        TCC
        TGT
        TTC
        */

        Queue<List<DnaQ>> ans = new ArrayDeque<List<DnaQ>>();
        Queue<List<UniPair<DnaQ>>> failed = new ArrayDeque<List<UniPair<DnaQ>>>();
        GlobalContext c = new GlobalContext(ans, failed, k, 11, 13, graph);

        DnaQBuilder leftBuilder = new DnaQBuilder();
        leftBuilder.append((byte) 3, (byte) 10);
        leftBuilder.append((byte) 3, (byte) 10);
        leftBuilder.append((byte) 0, (byte) 10);
        leftBuilder.append((byte) 1, (byte) 10);
        leftBuilder.append((byte) 0, (byte) 3);

        DnaQ left = leftBuilder.build();

        DnaQBuilder rightBuilder = new DnaQBuilder();
        rightBuilder.append((byte) 0, (byte) 10);
        rightBuilder.append((byte) 0, (byte) 10);
        rightBuilder.append((byte) 2, (byte) 10);
        rightBuilder.append((byte) 0, (byte) 10);
        rightBuilder.append((byte) 3, (byte) 10);

        DnaQ right = rightBuilder.build().complement();

        List<UniPair<DnaQ>> x = new ArrayList<UniPair<DnaQ>>();
        x.add(new UniPair<DnaQ>(left, right));

        FillingTask t = new FillingTask(c, x);
        t.run();

        List<DnaQ> ansList = ans.poll();
        // it's amibigous
        // assert ansList.size() == 1;
        // assert ansList.get(0).toString().equals("TAGGAATACA");
        // System.out.println("ok");
    }


}

