package ru.ifmo.genetics.tools.longReadsAssembler;

import org.apache.commons.configuration.Configuration;
import org.apache.commons.configuration.ConfigurationException;
import org.apache.commons.configuration.PropertiesConfiguration;
import org.apache.hadoop.mapred.OutputCollector;
import ru.ifmo.genetics.distributed.io.writable.DnaWritable;
import ru.ifmo.genetics.distributed.io.writable.Int128WritableComparable;
import ru.ifmo.genetics.dna.Dna;
import ru.ifmo.genetics.dna.DnaTools;
import ru.ifmo.genetics.statistics.Timer;
import ru.ifmo.genetics.tools.Util;
import ru.ifmo.genetics.tools.longReadsAssembler.layouter.*;
import ru.ifmo.genetics.tools.longReadsAssembler.overlaps.*;
import ru.ifmo.genetics.tools.longReadsAssembler.overlaps.overlapper.Overlapper;
import ru.ifmo.genetics.tools.longReadsAssembler.suffixArray.BucketsDivider;
import ru.ifmo.genetics.tools.longReadsAssembler.suffixArray.BucketsSorter;
import ru.ifmo.genetics.tools.longReadsAssembler.suffixArray.SuffixArray;

import java.io.IOException;
import java.io.PrintWriter;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;

/**
 * Assemblies all the given contigs in memory
 */
public class ContigsAssembler {
    Configuration globalConfig;

    public ContigsAssembler(Configuration globalConfig) {
        this.globalConfig = globalConfig;
    }


    /**
     * @param output if it is null then contigs will be written to the output file
     *               instead of being collected by OutputCollector
     */
    public void assemble(ArrayList<Dna> reads, OutputCollector<Int128WritableComparable, DnaWritable> output)
            throws IOException, InterruptedException {
        assemble(reads, output, globalConfig.getInt("min_reads_in_contig", 1), globalConfig);
    }

    /**
     * @param output if it is null then contigs will be written to the output file
     *               instead of being collected by OutputCollector
     */
    public void assemble(ArrayList<Dna> reads, OutputCollector<Int128WritableComparable, DnaWritable> output, int minReadsInContig)
            throws IOException, InterruptedException {
        assemble(reads, output, minReadsInContig, globalConfig);
    }


    /**
     * @param output              if it is null then contigs will be written to the output file
     *                            instead of being collected by OutputCollector
     */
    private void assemble(ArrayList<Dna> reads, OutputCollector<Int128WritableComparable, DnaWritable> output,
                         int minReadsInContig, Configuration config)
            throws IOException, InterruptedException {

        Timer timer = new Timer();
        Timer forAll = new Timer();

        if (reads.size() <= 1) {
            return;
        }


        System.out.println("Assembly starts");

        System.out.println("reads number: " + reads.size());
        long len = 0;
        for (Dna read : reads) {
            len += read.length();
        }
        System.out.println("total reads length = " + len);
        System.out.println("time = " + timer);
        System.out.println();


        // Thining reads - not used

        // Doubling reads
        timer.start();
        System.out.println("Doubling reads ...");
        reads = doubleReads(reads);
//        System.out.println("reads number: " + reads.calculateSize());
        System.out.println("time = " + timer);
//        System.out.println("-----------------------------------------");
        System.out.println();

        // Gluing reads into one string
        timer.start();
        System.out.println("Gluing reads into one string ...");
        GluedDnasString fullString = glueReads(reads);
        System.out.println("fullString length: " + fullString.length());
        System.out.println("time = " + timer);
//        System.out.println("-----------------------------------------");
        System.out.println();

        // SA buckets dividing
        timer.start();
        System.out.println("SA buckets dividing ...");
        BucketsDivider divider = new BucketsDivider(config.subset("overlapper"));
        SuffixArray[] sas = divider.divideIntoBuckets(fullString);
        divider = null;
        System.out.println("buckets number: " + sas.length);
        System.out.println("time = " + timer);
//        System.out.println("-----------------------------------------");
        System.out.println();

        // SA buckets sorting
        timer.start();
        System.out.println("SA buckets sorting ...");
        BucketsSorter sorter = new BucketsSorter(config.subset("overlapper"));
        sorter.sortAll(sas);
        sorter = null;
        System.out.println("time = " + timer);
//        System.out.println("-----------------------------------------");
        System.out.println();

        // Finding overlaps
        timer.start();
        System.out.println("Finding overlaps ...");
        Overlapper overlapper = new Overlapper(config.subset("overlapper"), fullString);
        Overlaps overlaps = new Overlaps(reads, config.subset("overlapper"));
        overlapper.findOverlaps(sas, overlaps);
        overlapper = null;
//        System.out.println("overlaps number: " + overlaps.calculateSize());
        System.out.println("time = " + timer);
//        System.out.println("-----------------------------------------");
        System.out.println();

        // Merging overlaps - not used

//        // Adding non-detected overlaps
//        timer.start();
//        System.out.println("Adding non-detected overlaps ...");
//        overlaps.setChanges();
//        OverlapsSaturator saturator = new OverlapsSaturator(config.subset("overlapper"));
//        saturator.addNonDetectedOverlaps(overlaps, reads);
//        saturator = null;
//        System.out.println("time = " + timer);
////        System.out.println("-----------------------------------------");
//        System.out.println();
//
//        // Removing false overlaps
//        timer.start();
//        System.out.println("Removing false overlaps ...");
//        overlaps.setChanges();
//        OverlapsWeeder weeder = new OverlapsWeeder(config.subset("layouter"));
//        weeder.removeFalseOverlaps(overlaps, reads);
//        weeder = null;
//        System.out.println("time = " + timer);
////        System.out.println("-----------------------------------------");
//        System.out.println();

        // Optimizing overlaps
        timer.start();
        System.out.println("Optimizing overlaps ...");
        OverlapsOptimizer.optimizeOverlaps(overlaps);
        System.out.println("time = " + timer);
//        System.out.println("-----------------------------------------");
        System.out.println();

        // Layouting
        timer.start();
        System.out.println("Layouting ...");
        Layouter layouter = new Layouter(config.subset("layouter"));
        layouter.loadOverlaps(reads, overlaps);
        layouter.mergeGraph();
//        layouter.removeTips();
        boolean[] goodReads = layouter.findGoodReads();
        System.out.println("time = " + timer);
//        System.out.println("-----------------------------------------");
        System.out.println();

        // Layout-Consensus
        timer.start();
        System.out.println("Layout-Consensus ...");
        String outputFile = config.getString("output");
        ConsensusLayoutWriter writer = (output == null) ? new SimpleConsensusLayoutWriter(0, reads, minReadsInContig, new PrintWriter(outputFile))
                : new ConsensusLayoutCollectorWriter(0, reads, minReadsInContig, output);

        layouter.dumpResult(goodReads, writer);

        System.out.println("time = " + timer);
//        System.out.println("-----------------------------------------");
        System.out.println();

        // Statistic
        System.out.println("Statistic:");
        System.out.println(writer.getStat());

        System.out.println("All time = " + forAll);
//        System.out.println("-----------------------------------------");
        System.out.println();
    }

    public static ArrayList<Dna> doubleReads(ArrayList<Dna> reads) {
        ArrayList<Dna> dReads = new ArrayList<Dna>();
        for (Dna dna : reads) {
            dReads.add(dna);
            dReads.add(dna.reverseComplement());
        }
        return dReads;
    }

    /**
     * Glues all reads in one string, formed as 'read1$read2$...$readN$'
     */
    public static GluedDnasString glueReads(ArrayList<Dna> reads) {
        long size = 0;
        for (Dna dna : reads) {
            size += dna.length() + 1;
        }

        GluedDnasString res = new GluedDnasString(size);

        long pos = 0;
        for (Dna dna : reads) {
            for (int i = 0; i < dna.length(); i++) {
                res.set(pos, GluedDnasString.transform[DnaTools.toChar(dna.nucAt(i))]);
                pos++;
            }
            res.set(pos, GluedDnasString.$index);
            pos++;
        }

        return res;
    }


    public static void main(String[] args) throws IOException, ConfigurationException, InterruptedException, NoSuchAlgorithmException {
        Configuration config = new PropertiesConfiguration(args[0]);
        new ContigsAssembler(config).run();
    }

    public void run() throws IOException, InterruptedException, NoSuchAlgorithmException {
        String readsFile = globalConfig.getString("reads");
        ArrayList<Dna> reads = ReadsPlainReader.loadReadsAndAddRC(readsFile);

        assemble(reads, null);
    }

}
