package ru.ifmo.genetics.tools.longReadsAssembler.layouter;

import org.apache.commons.cli.*;
import org.apache.commons.configuration.Configuration;
import org.apache.commons.configuration.ConfigurationException;
import org.apache.commons.configuration.PropertiesConfiguration;
import ru.ifmo.genetics.dna.Dna;
import ru.ifmo.genetics.tools.Util;
import ru.ifmo.genetics.tools.longReadsAssembler.ReadsPlainReader;
import ru.ifmo.genetics.tools.longReadsAssembler.overlaps.Overlaps;
import ru.ifmo.genetics.tools.longReadsAssembler.overlaps.OverlapsList;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.PriorityQueue;

import static ru.ifmo.genetics.tools.longReadsAssembler.overlaps.OverlapsList.Edge;

public class Layouter {

    private String readsFile = null;
    private String outputFile;
    private String overlapsFile;
    private int readsNumber = -1;
    // final char[] DNA = "ATGC".toCharArray();

    private Configuration config;

    public Layouter(Configuration config) {
        readsFile = config.getString("reads");
        outputFile = config.getString("output");
        overlapsFile = config.getString("overlaps");
        this.config = config;
    }


    private Overlaps<Dna> allOverlaps;


    boolean[] was;
    private ArrayList<Dna> reads;

    /**
     * Returns true is i was in tip and it was removed
     */
    private boolean removeTipsIfAny(int i, int depth) {
        if (depth < 0) {
            return false;
        }
        OverlapsList tempList = new OverlapsList();
        if (allOverlaps.getBackwardOverlaps(i, tempList).size() > 1) {
            return false;
        }

        allOverlaps.getForwardOverlaps(i, tempList);
        for (int j = 0; j < tempList.size(); ++j) {
            if (!removeTipsIfAny(tempList.getTo(j), depth - 1))
                return false;
        }

        removeNode(i);
        return true;
    }

    private boolean removeTipsIfAny(int i) {
        return removeTipsIfAny(i, 5);
    }

    private void removeNode(int i) {
        OverlapsList tempList = allOverlaps.getForwardOverlaps(i, new OverlapsList());
        for (int j = 0; j < tempList.size(); ++j) {
            int to = tempList.getTo(j);
            int shift = tempList.getCenterShift(j);
            allOverlaps.removeOverlap(i, to, shift);
        }

        tempList = allOverlaps.getBackwardOverlaps(i, tempList);

        for (int j = 0; j < tempList.size(); ++j) {
            int to = tempList.getTo(j);
            int shift = tempList.getCenterShift(j);
            allOverlaps.removeOverlap(to, i, -shift);
        }
        allOverlaps.markReadRemoved(i);
    }

    private void run() throws IOException, InterruptedException {
    	
        loadOverlaps();

        mergeGraph();
        removeTips();
        removeNonMinimalOverlaps();
        allOverlaps.printToFile("overlaps.layout");
        boolean[] goodReads = findGoodReads();

        dumpResult(goodReads, new SimpleLayoutWriter(new PrintWriter(outputFile)));

    }

    private void removeNonMinimalOverlaps() {
        Overlaps<Dna> newOverlaps = new Overlaps<Dna>(reads, config);
        
        OverlapsList tempList = new OverlapsList();
        for (int i = 0; i < readsNumber; ++i) {
            if (allOverlaps.isReadRemoved(i)) {
                newOverlaps.markReadRemoved(i);
                continue;
            }
            allOverlaps.getForwardOverlaps(i, tempList);

            int minShift = -1;
            int minJ = -1;
            for (int j = 0; j < tempList.size(); ++j) {
                int jTo = tempList.getTo(j);
                int jCenterShift = tempList.getCenterShift(j);

                if (j == 0 || jCenterShift < minShift) {
                    minShift = jCenterShift;
                    minJ = j;
                }
            }

            if (minJ != -1) {
                newOverlaps.addOverlap(i, tempList.getTo(minJ), tempList.getCenterShift(minJ));
            }

        }
        allOverlaps = newOverlaps;

    }

    public boolean[] findGoodReads() {
        OverlapsList tempList = new OverlapsList();
		boolean[] goodReads = new boolean[readsNumber];

        int badsNumber = readsNumber;
        int lonelyNumber = 0;
        boolean[] lonely = new boolean[readsNumber];
        for (int i =  0; i < readsNumber; ++i) {
            if (allOverlaps.isReadRemoved(i)) {
                continue;
            }
            int s1 = allOverlaps.getForwardOverlaps(i, tempList).size();
            if (s1 > 1)
                continue;

            int s2 = allOverlaps.getBackwardOverlaps(i, tempList).size();
            if (s2 > 1)
                continue;

            lonely[i] = s1 == 0 && s2 == 0;
            if (lonely[i]) {
                lonelyNumber++;
            }
            goodReads[i] = true;
//            goodReads[i] = !lonely[i];
            --badsNumber;
        }

//        System.out.println("good reads found, bads - " + badsNumber);
//        System.out.println("lonely - " + lonelyNumber);
		return goodReads;
	}

	public int removeTips() {
        OverlapsList tempList = new OverlapsList();

        int tipsRemoved = 0;
        for (int i = 0; i < readsNumber; ++i) {
            if (allOverlaps.isReadRemoved(i)) {
                continue;
            }
            if (allOverlaps.getForwardOverlaps(i, tempList).size() > 1) {
                for (int j = 0; j < tempList.size(); ++j) {
                    if (removeTipsIfAny(tempList.getTo(j))) {
                        tipsRemoved++;
                    }
                }
            }
        }

        return tipsRemoved;

//        System.out.println(tipsRemoved + " tips removed");
	}

	public void mergeGraph() throws InterruptedException {
//    	System.out.println("overlaps before merging graph: " + allOverlaps.calculateSize());
		OverlapsList tempList = new OverlapsList();
        int tries = 0;
        for (int i = 0; i < readsNumber; ++i) {
        	/*
            if ((i & (i - 1)) == 0) {
                System.err.println(i + "/" + readsCount);
            }
            */
            if (allOverlaps.isReadRemoved(i)) {
                continue;
            }
            if (allOverlaps.getForwardOverlaps(i, tempList).size() > 1) {
                tryMergePaths(i, 1000);
                tries++;
                /*
                if ((tries & (tries - 1)) == 0) {
                    System.err.println(tries + " tries");
                }
                */
            }
        }
//        System.out.println("overlaps after merging graph: " + allOverlaps.calculateSize());

	}

	public void dumpResult(boolean[] goodReads, LayoutWriter writer)
			throws FileNotFoundException, IOException {
//        PrintWriter layoutOverlaps = new PrintWriter("overlaps.layout");
        
//        System.err.println("readsCount = " + readsCount);

        OverlapsList tempList = new OverlapsList();
        was = new boolean[readsNumber];

        for (int i = 0; i < readsNumber; ++i) {
            if (allOverlaps.isReadRemoved(i)) {
                continue;
            }
        	ArrayList<Integer> readsInCurrentContig = new ArrayList<Integer>();
            if (was[i]) {
                continue;
            }

            if (!goodReads[i]) {
//                for (int j = 0; j < allOverlaps.getForwardOverlaps(i, tempList).calculateSize(); ++j) {
//                    layoutOverlaps.println(i + " " + tempList.getTo(j) + " " + tempList.getCenterShift(j));
//                }
                continue;
            }

            int start = i;
//            System.err.println("going back from " + i);
            while (true) {
                allOverlaps.getBackwardOverlaps(start, tempList);
                if (tempList.size() == 0) {
                    break;
                }

                int newStart = tempList.getTo(0);

                if (!goodReads[newStart]) {
                    break;
                }

                start = newStart;
                if (start == i) {
                    break;
                }
            }
//            System.err.println("going forward from " + start);
            int shift = 0;
            int cur = start;
            int end = start;
            int endShift = shift;
            while (!was[cur]) {
            	writer.addLayout(cur, shift);
                readsInCurrentContig.add(cur);
                end = cur;
                endShift = shift;
                was[cur] = true;

                allOverlaps.getForwardOverlaps(cur, tempList);

                if (tempList.size() == 0) {
                    break;
                }

                int newCur = tempList.getTo(0);

                if (!goodReads[newCur]) {
                    break;
                }

                shift += allOverlaps.centerShiftToBeginShift(cur, tempList.getTo(0), tempList.getCenterShift(0));
                cur = newCur;
            }
//            if (allOverlaps.getForwardOverlaps(end, tempList).calculateSize() == 0) {
//                if (allOverlaps.getBackwardOverlaps(start, tempList).calculateSize() != 0)
//                layoutOverlaps.println(start + " " + end + " " + endShift);
//            }
//            allOverlaps.getForwardOverlaps(end, tempList);
//            for (int j = 0; j < tempList.calculateSize(); ++j) {
//                layoutOverlaps.println(start + " " + tempList.getTo(j) + " " + (tempList.getCenterShift(j) + endShift));
//            }
            writer.flush();
            for (Integer r: readsInCurrentContig) {
                was[r.intValue() ^ 1] = true;
            }
        }

        for (int i = 0; i < readsNumber; ++i) {
            if (allOverlaps.isReadRemoved(i)) {
                continue;
            }
            if (was[i]) {
                continue;
            }
            writer.addLayout(i, 0);
            writer.flush();
            was[i] = true;
            was[i ^ 1] = true;
        }
        writer.close();
//        layoutOverlaps.close();
	}

    public void loadOverlaps(ArrayList<Dna> reads, Overlaps allOverlaps) throws InterruptedException {
		this.reads = reads;
		readsNumber = reads.size();

		this.allOverlaps = allOverlaps;
		this.allOverlaps.sortAll();
	}


	public void loadOverlaps() throws IOException, InterruptedException {
        reads = ReadsPlainReader.loadReadsAndAddRC(readsFile);
        readsNumber = reads.size();
        System.err.println(readsNumber + " reads loaded");

        allOverlaps = new Overlaps<Dna>(reads, overlapsFile, config);

        allOverlaps.sortAll();
        System.err.println("overlaps loaded");
	}

    private void removeOverlap(Edge from, Edge to) {
        allOverlaps.removeOverlap(from.to, to.to, to.centerShift - from.centerShift);
    }

    private void addOverlap(Edge from, Edge to) {
        allOverlaps.addOverlap(from.to, to.to, to.centerShift - from.centerShift);
    }

    private void mergeBackPaths(Edge end, Edge to1, Edge to2, HashMap<Edge, Edge> prevs) {
//        System.err.println(end + " " + to1 + " " + to2);
        // to1 is connected to end
        int z = to2.compareTo(to1);
        if (to1.equals(to2)) {
            return;
        }

        try {
            if (allOverlaps.isWellOriented(to2.to, to1.to, to1.centerShift - to2.centerShift)) {
                Edge newTo2 = prevs.get(to1);
                removeOverlap(newTo2, to1);
                addOverlap(to2, to1);
                prevs.put(to1, to2);
                mergeBackPaths(to1, to2, newTo2, prevs);
            } else {
                Edge newTo1 = prevs.get(to2);
                removeOverlap(to1, end);
                addOverlap(to2, end);
                prevs.put(end, to2);
                mergeBackPaths(to2, newTo1, to1, prevs);
            }
        } catch (StackOverflowError e) {
            System.err.println(end + " " + to1 + " " + to2);
            throw e;
        }
    }

    private void tryMergePaths(int from, int depth) {
//        System.err.println("merging from " + from);
        PriorityQueue<Edge> queue = new PriorityQueue<Edge>();
        HashMap<Edge, Edge> prevs = new HashMap<Edge, Edge>();
        HashSet<Integer> visited = new HashSet<Integer>();
        queue.add(new Edge(from, 0));
        prevs.put(new Edge(from, 0), null);
        boolean firstly = true;
//        int merges = 0;
        OverlapsList tempList = new OverlapsList();
        while (true) {
            if (!firstly && queue.size() == 1) {
                break;
            }

            firstly = false;


            Edge u = queue.poll();
            
            if (u.centerShift > depth) {
                break;
            }

            if (!visited.add(u.to)) {
                continue;
            }

            {
                int visitedSize = prevs.size();
                if ((visitedSize & 65536) != 0) {
                    break;
                }
            }

            allOverlaps.getForwardOverlaps(u.to, tempList);

            for (int i = 0 ; i < tempList.size(); ++i) {
                Edge v = tempList.get(i);
                v.centerShift += u.centerShift;

                if (prevs.containsKey(v)) {
//                    System.err.println("merging to " + v);
                    mergeBackPaths(v, prevs.get(v), u, prevs);
                    /*
                    ArrayList<Edge> firstPath = getBackPath(prevs.get(v), prevs);
                    ArrayList<Edge> secondPath = getBackPath(u, prevs);
                    firstPath.add(v);
                    secondPath.add(v);
                    breakInPath(firstPath, prevs);
                    breakInPath(secondPath, prevs);


                    ArrayList<Edge> newPath = mergeLists(firstPath, secondPath);
                    joinInPath(newPath, prevs);
                    ++merges;
                    */

//                    System.err.println(firstPath);
//                    System.err.println(secondPath);
//                    System.err.println(newPath);
                    continue;
                }

                prevs.put(v, u);
                queue.add(v);
            }
        }
//        System.err.println(merges + " merges done");

    }


    public static void main(String[] args) throws IOException, InterruptedException {
        Options options = new Options();

        options.addOption("h", "help", false, "prints this message");
        options.addOption("c", "config", true,  "sets the config file name, default to config.properties");
        options.addOption("o", "output", true,  "sets the ouput file name");
        options.addOption("O", "overlaps", true,  "sets the overlaps file name");
        options.addOption("r", "reads", true,  "sets the reads file name");
        options.addOption("n", "reads-number", true,  "sets the number of reads");

        CommandLineParser parser = new PosixParser();
        CommandLine cmd;
        try {
            cmd = parser.parse(options, args);
        } catch (ParseException e) {
            e.printStackTrace(System.err);
            return;
        }

        if (cmd.hasOption("help")) {
            new HelpFormatter().printHelp("layout", options);
            return;
        }

        String configFileName = cmd.getOptionValue("config", "config.properties");
        Configuration config;
        try {
            config = new PropertiesConfiguration(configFileName).subset("layouter");
        } catch (ConfigurationException e) {
            e.printStackTrace(System.err);
            return;
        }

        Util.addOptionToConfig(cmd, config, "output");
        Util.addOptionToConfig(cmd, config, "reads");
        Util.addOptionToConfig(cmd, config, "overlaps");
        Util.addOptionToConfig(cmd, config, "reads-number");

        new Layouter(config).run();
    }

}
