package ru.ifmo.genetics.tools.longReadsAssembler.overlaps;

import org.apache.commons.configuration.Configuration;
import ru.ifmo.genetics.dna.DnaTools;
import ru.ifmo.genetics.dna.DnaView;
import ru.ifmo.genetics.dna.LightDna;
import ru.ifmo.genetics.tools.Util;
import ru.ifmo.genetics.tools.executors.BlockingThreadPoolExecutor;
import ru.ifmo.genetics.tools.longReadsAssembler.ReaderInSmallMemory;
import ru.ifmo.genetics.utils.Misc;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;


public class Overlaps<T extends LightDna> {
    OverlapsList[] overlaps;
    public ArrayList<T> reads;
    
    final int readsNumber;
    final int availableProcessors;

    
    public static <T extends LightDna> boolean assertReadsAreWithRCCopy(ArrayList<T> reads) {
        for (int i = 0; i < reads.size(); i += 2) {
            assert DnaTools.equals(DnaView.rcView(reads.get(i)), reads.get(i + 1));
        }
        return true;
    }

    protected Overlaps(ArrayList<T> reads, Configuration config, boolean createOverlapLists) {
        assert assertReadsAreWithRCCopy(reads);

        this.reads = reads;
        readsNumber = reads.size();
        
        overlaps = new OverlapsList[readsNumber];
        
        availableProcessors = config.getInt("available_processors");

        if (createOverlapLists) {
            for (int i = 0; i < readsNumber; ++i) {
                overlaps[i] = new OverlapsList();
            }
        }
    }

    public Overlaps(ArrayList<T> reads, Configuration config) {
        this(reads, config, true);
    }

    public Overlaps(ArrayList<T> reads, String overlapsFile, Configuration config) throws IOException, InterruptedException {
        this(reads, config, false);

        // loading
        long inputOverlapsNumber = Misc.linesNumber(overlapsFile);
        System.err.println(inputOverlapsNumber + " overlaps to read");

        for (int i = 0; i < readsNumber; ++i) {
            overlaps[i] = new OverlapsList((int)(inputOverlapsNumber / readsNumber / 2) + 1);
        }

        ReaderInSmallMemory overlapsReader = new ReaderInSmallMemory(new File(overlapsFile));
        boolean rawOverlaps = overlapsFile.endsWith(".raw");
        int taskSize = 1 << 14;
        BlockingThreadPoolExecutor executor = new BlockingThreadPoolExecutor(availableProcessors);
        while (true) {
            FullOverlapsList task = new FullOverlapsList(taskSize);
            boolean finished = false;
            for (int i = 0; i < taskSize; ++i) {
                int from = overlapsReader.readInteger();
                if (from == -1) {
                    finished = true;
                    break;
                }
                int to = overlapsReader.readInteger();
                int shift = overlapsReader.readInteger();
                task.add(from, to, shift);
            }

            executor.blockingExecute(new AddingTask(this, task, rawOverlaps));
            if (finished)
                break;
        }

        overlapsReader.close();
        executor.shutdownAndAwaitTermination();
    }

    public boolean isReadRemoved(int i) {
        assert (overlaps[i] == null) == (overlaps[i ^ 1] == null);
        return overlaps[i] == null;
    }
    
    public void markReadRemoved(int i) {
        overlaps[i] = null;
        overlaps[i ^ 1] = null;
    }

    /**
     *
     * @param x
     * @return x cyclic shifted right by 1 bit
     */
    private static int getKey(int x) {
        if ((x & 1) == 1) {
            return -x - 1;
        } else {
            return x;
        }
    }
    
    public static int compare(int read1, int centerShift1, int read2, int centerShift2) {
        int cmp = centerShift1 - centerShift2;
        if (cmp != 0) {
            return cmp;
        }

        return Util.compare(getKey(read1), getKey(read2));
    }

    /**
     * Checks if overlap is oriented from from to to
     *
     * Orientations from from to to means that a center of the first read
     * goes earlier than a center of the second. If they are on the same
     * position then cyclic-shifted-right-by-1-bit ids are compared.
     *
     * @param from
     * @param to
     * @param centerShift
     * @return
     */
    public boolean isWellOriented(int from, int to, int centerShift) {
        return compare(from, 0, to, centerShift) <= 0;
    }

    /**
     * Add forward and reverse overlaps. Not thread-safe.
     *
     * Overlap need not to be well-oriented.
     *
     * @param from
     * @param to
     * @param centerShift
     * @return number of added overlaps, usually it's 0 or 2, but sometimes mb 1
     */
    public int addOverlap(int from, int to, int centerShift) {
        if (!isWellOriented(from, to, centerShift)) {
            return addOverlap(to, from, -centerShift);
        }

        if (overlaps[from].contains(to, centerShift)) {
            return 0;
        }

        overlaps[from].add(to, centerShift);
        int res = 1;

        if (from != (to ^ 1)) {
            int cfrom = to ^ 1;
            int cto = from ^ 1;
            assert isWellOriented(cfrom, cto, centerShift);
            overlaps[cfrom].add(cto, centerShift);
            res++;
        }

        return res;
    }


    public boolean containsOverlap(int from, int to, int centerShift) {
        if (!isWellOriented(from, to, centerShift)) {
            return containsOverlap(to, from, -centerShift);
        }

        return overlaps[from].contains(to, centerShift);
    }

    /**
     * Add forward and reverse overlaps. Thread-safe.
     *
     * Overlap need not to be well-oriented.
     *
     * @param from
     * @param to
     * @param centerShift
     * @return number of added overlaps, usually it's 0 or 2, but sometimes mb 1
     */
    public int addOverlapWithSync(int from, int to, int centerShift) {
        int x = from & (from ^ 1);
        int y = to & (to ^ 1);
        synchronized (overlaps[Math.min(x, y)]) {
            synchronized (overlaps[Math.max(x, y)]) {
                return addOverlap(from, to, centerShift);
            }
        }

    }

    public int addRawOverlapWithSync(int from, int to, int beginShift) {
        return addOverlapWithSync(from, to, beginShiftToCenterShift(from, to, beginShift));
    }

    /**
     * Remove forward and reverse overlaps. Not thread-safe. Complement to addOverlap()
     *
     * Overlap need not to be well-oriented.
     *
     * @param from
     * @param to
     * @param centerShift
     * @return number of added overlaps, usually it's 0 or 2, but sometimes mb 1
     */
    public int removeOverlap(int from, int to, int centerShift) {
        if (!isWellOriented(from, to, centerShift)) {
            return removeOverlap(to, from, -centerShift);
        }

        if (!overlaps[from].remove(to, centerShift)) {
            return 0;
        }

        int res = 1;

        if (from != (to ^ 1)) {
            int cfrom = to ^ 1;
            int cto = from ^ 1;
            assert isWellOriented(cfrom, cto, centerShift);
            overlaps[cfrom].remove(cto, centerShift);
            res++;
        }

        return res;
    }

    /**
     * Adds all overlaps from from to list.
     *
     * @param from
     * @param result list to add to
     * @return
     */
    public OverlapsList getForwardOverlaps(int from, OverlapsList result) {
        result.clear();
//        return getExtendedOverlaps(from, result);
        result.addAll(overlaps[from]);
        return result;
    }

    /**
     * Adds all back overlaps from from to list.
     *
     * @param from
     * @param result list to add to
     * @return
     */
    public OverlapsList getBackwardOverlaps(int from, OverlapsList result) {
        result.clear();
        OverlapsList rcOverlaps = overlaps[from ^ 1];
        result.ensureCapacity(rcOverlaps.size());

        for (int i = 0; i < rcOverlaps.size(); ++i) {
            int to = rcOverlaps.getTo(i) ^ 1;
            int shift = -rcOverlaps.getCenterShift(i);
            result.add(to, shift);
        }
        return result;
    }

    public void sortAll() throws InterruptedException {
        BlockingThreadPoolExecutor sortExecutor = new BlockingThreadPoolExecutor(availableProcessors);
        int taskSize = readsNumber / availableProcessors + 1;
        for (int i = 0; i < readsNumber; i += taskSize) {
            sortExecutor.blockingExecute(new SortingTask(this, i, Math.min(i + taskSize, readsNumber)));
        }
        sortExecutor.shutdownAndAwaitTermination();
    }

    public void printToFile(String outputFile) throws FileNotFoundException {
        PrintWriter out = new PrintWriter(outputFile);
        for (int i = 0; i < readsNumber; ++i) {
            if (overlaps[i] == null) {
                out.println(i + " -1 0");
                continue;
            }
            for (int j = 0; j < overlaps[i].size(); ++j) {
                out.println(i + " " + overlaps[i].getTo(j) + " " + overlaps[i].getCenterShift(j));
            }
        }
        out.close();
    }

    public int beginShiftToCenterShift(int from, int to, int beginShift) {
        int length1 = reads.get(from).length();
        int length2 = reads.get(to).length();

        return 2 * beginShift + length2 - length1;
    }

    public int centerShiftToBeginShift(int from, int to, int centerShift) {
        int length1 = reads.get(from).length();
        int length2 = reads.get(to).length();
        int doubleBeginShift = centerShift + length1 - length2;
        assert doubleBeginShift % 2 == 0;

        return doubleBeginShift / 2;
    }

}
