package ru.ifmo.genetics.tools.longReadsAssembler.overlaps;

import org.apache.commons.cli.*;
import org.apache.commons.configuration.Configuration;
import org.apache.commons.configuration.ConfigurationException;
import org.apache.commons.configuration.PropertiesConfiguration;
import ru.ifmo.genetics.dna.Dna;
import ru.ifmo.genetics.statistics.Timer;
import ru.ifmo.genetics.tools.Util;
import ru.ifmo.genetics.tools.executors.BlockingThreadPoolExecutor;
import ru.ifmo.genetics.tools.longReadsAssembler.ReadsPlainReader;

import java.io.IOException;
import java.util.ArrayList;


public class OverlapsOptimizer {
    String readsFile;
    String overlapsFile;
    String outputFile;

    Configuration config;

    public OverlapsOptimizer(Configuration config) {
        readsFile = config.getString("reads");
        overlapsFile = config.getString("overlaps");
        outputFile = config.getString("output");
        this.config = config;
    }


    public static void main(String[] args) throws IOException, InterruptedException {
        Options options = new Options();

        options.addOption("h", "help", false, "prints this message");
        options.addOption("c", "config", true,  "sets the config file name, default to config.properties");
        options.addOption("o", "output", true,  "sets the ouput file name");
        options.addOption("O", "overlaps", true,  "sets the overlaps file name");
        options.addOption("r", "reads", true,  "sets the reads file name");

        CommandLineParser parser = new PosixParser();
        CommandLine cmd;
        try {
            cmd = parser.parse(options, args);
        } catch (ParseException e) {
            e.printStackTrace(System.err);
            return;
        }

        if (cmd.hasOption("help")) {
            new HelpFormatter().printHelp("overlap", options);
            return;
        }

        String configFileName = cmd.getOptionValue("config", "config.properties");
        Configuration config;
        try {
            config = new PropertiesConfiguration(configFileName).subset("layouter");
        } catch (ConfigurationException e) {
            e.printStackTrace(System.err);
            return;
        }

        Util.addOptionToConfig(cmd, config, "output");
        Util.addOptionToConfig(cmd, config, "reads");
        Util.addOptionToConfig(cmd, config, "overlaps");

        new OverlapsOptimizer(config).run();
    }

    public void run() throws IOException, InterruptedException {
        Overlaps allOverlaps = loadOverlaps();

        allOverlaps.sortAll();
        System.err.println("overlaps sorted");

        Timer timer = new Timer();
        optimizeOverlaps(allOverlaps);
        System.err.println("optimizing overlaps took " + timer);

        allOverlaps.printToFile(outputFile);
    }

    private Overlaps loadOverlaps() throws IOException, InterruptedException {
        ArrayList<Dna> reads = ReadsPlainReader.loadReadsAndAddRC(readsFile);

        System.err.println(reads.size() + " reads loaded");
        Overlaps allOverlaps = new Overlaps(reads, overlapsFile, config);
        System.err.println("overlaps loaded");
        return allOverlaps;
    }


    public static void optimizeOverlaps(Overlaps overlaps) throws InterruptedException {
//        System.out.println("overlaps before optimising: " + calculateSize());
        OverlapsList[] newOverlaps = new OverlapsList[overlaps.readsNumber];
        BlockingThreadPoolExecutor executor = new BlockingThreadPoolExecutor(overlaps.availableProcessors);
        int taskSize = overlaps.readsNumber / overlaps.availableProcessors + 1;
        for (int i = 0; i < overlaps.readsNumber; i += taskSize) {
            executor.blockingExecute(new OptimizingTask(overlaps, i, Math.min(i + taskSize, overlaps.readsNumber), newOverlaps));
        }
        executor.shutdownAndAwaitTermination();
        overlaps.overlaps = newOverlaps;
//        System.out.println("overlaps after optimising: " + calculateSize());

    }

    static class OptimizingTask implements Runnable {
        Overlaps overlaps;
        final int begin;
        final int end;

        OverlapsList oldOverlaps = new OverlapsList();
        OverlapsList ends = new OverlapsList();
        OverlapsList.OverlapsSortTraits sortTraits;
        OverlapsList[] newOverlaps;

        OptimizingTask(Overlaps overlaps, int begin, int end, OverlapsList[] newOverlaps) {
            this.overlaps = overlaps;
            this.begin = begin;
            this.end = end;
            this.newOverlaps = newOverlaps;
             sortTraits = new OverlapsList.OverlapsSortTraits();
        }

        void optimize(int i) {
            overlaps.getForwardOverlaps(i, oldOverlaps);
            oldOverlaps.sort(sortTraits);

//            System.err.println(i + ":");
//            System.err.println("extended");
//            System.err.println(oldOverlaps);
//            System.err.println("before");
//            System.err.println(overlaps[i]);

            ends.clear();


//            int prevSize = overlaps.overlaps[i].size();

            for (int j = oldOverlaps.size() - 1; j >= 0; --j) {
                int eto = oldOverlaps.getTo(j);
                int eshift = oldOverlaps.getCenterShift(j);
                if (eto == i && eshift == 0) {
                    continue;
                }

                for (int endIt = 0; endIt < ends.size(); ++endIt) {
                    int prevto = ends.getTo(endIt);
                    int prevshift = ends.getCenterShift(endIt);

                    int seqShift = (eshift - prevshift);


                    if (overlaps.containsOverlap(prevto, eto, seqShift)) {
                        ends.remove(endIt--);
                    }
                }

                ends.add(eto, eshift);
            }
//            int afterSize = ends.size();
            newOverlaps[i] = new OverlapsList(ends);
            newOverlaps[i].sort(sortTraits);

//            System.err.println("after");
//            System.err.println(overlaps[i]);
//            System.err.println(i + ": " + prevSize + " before, " + afterSize + " after");


        }

        @Override
        public void run() {
            for (int i = begin; i < end; ++i) {
                if (overlaps.overlaps[i] == null) {
                    continue;
                }
                optimize(i);
            }
        }
    }

}
