package ru.ifmo.genetics.tools.longReadsAssembler.overlaps.overlapper;

import static ru.ifmo.genetics.tools.longReadsAssembler.GluedDnasString.$index;

import org.apache.commons.cli.*;
import org.apache.commons.configuration.Configuration;
import org.apache.commons.configuration.ConfigurationException;
import org.apache.commons.configuration.PropertiesConfiguration;

import ru.ifmo.genetics.statistics.Timer;
import ru.ifmo.genetics.tools.DedicatedWriter;
import ru.ifmo.genetics.tools.IntComparator;
import ru.ifmo.genetics.tools.Util;
import ru.ifmo.genetics.tools.executors.PatientExecutorService;
import ru.ifmo.genetics.tools.longReadsAssembler.GluedDnasString;
import ru.ifmo.genetics.tools.longReadsAssembler.overlaps.Overlaps;
import ru.ifmo.genetics.tools.longReadsAssembler.suffixArray.BucketsDivider;
import ru.ifmo.genetics.tools.longReadsAssembler.suffixArray.SuffixArray;
import ru.ifmo.genetics.tools.longReadsAssembler.suffixArray.BucketsSorter;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;

public class Overlapper implements Runnable {
    final String bucketsDir;
    
    final int numberOfErrors;
    final int errorsWindowSize;
    final int minOverlap;
    final int bucketCharsNumber;
    
    final int availableProcessors;
    
    final String output;

    public static void main(String[] args) throws IOException {
        Configuration config = getOverlapperConfiguration(args);
        if (config == null) {
            return;
        }

        Overlapper overlapper = new Overlapper(config, null);
        overlapper.run();
    }
    
    static Configuration getOverlapperConfiguration(String[] args) {
        Options options = new Options();
        options.addOption("h", "help", false, "prints this message");
        options.addOption("c", "config", true,  "sets the config file, default to config.properties");
        options.addOption("o", "output", true, "sets the output file");
        options.addOption("r", "reads", true, "sets the file with glued reads");
        options.addOption("b", "buckets-dir", true, "sets the buckets directory");
        options.addOption("B", "bucket-chars-number", true, "sets the prefix calculateSize for buckets");

        CommandLineParser parser = new PosixParser();
        CommandLine cmd;
        try {
            cmd = parser.parse(options, args);
        } catch (ParseException e) {
            e.printStackTrace(System.err);
            return null;
        }
        if (cmd.hasOption("help")) {
            new HelpFormatter().printHelp("overlap", options);
            return null;
        }

        String configFileName = cmd.getOptionValue("config", "config.properties");
        Configuration config;
        try {
            config = new PropertiesConfiguration(configFileName).subset("overlapper");
        } catch (ConfigurationException e) {
            e.printStackTrace(System.err);
            return null;
        }

        Util.addOptionToConfig(cmd, config, "output");
        Util.addOptionToConfig(cmd, config, "reads", "reads_for_overlaps");
        Util.addOptionToConfig(cmd, config, "buckets-dir");
        Util.addOptionToConfig(cmd, config, "bucket-chars-number");
        
        return config;
    }

    @Override
    public void run() {
        try {
        	Timer t = new Timer();
            findOverlaps();
            System.err.println("numberOfErrors = " + numberOfErrors);
            System.err.println("Make approximate simple overlaps in " + t);
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    
    /**
     * <code>readBegin[i]</code> = position in <code>text</code> where read number <code>i</code> begins.
     */
    private long[] readBegin;
    private int readsNumber;

    /**
     * String with all reads, formed as 'read1$read2$...$readN$'.
     */
    private final GluedDnasString text;


    public Overlapper(Configuration config, GluedDnasString text) throws IOException {
        bucketsDir = config.getString("buckets_dir");
        numberOfErrors = config.getInt("number_of_errors");
        errorsWindowSize = config.getInt("errors_window_size");
        minOverlap = config.getInt("min_overlap");
        bucketCharsNumber = config.getInt("bucket_chars_number");
        availableProcessors = config.getInt("available_processors");
        output = config.getString("output");
        
        if (text != null) {
            this.text = text;
        } else {
            this.text = new GluedDnasString(new File(config.getString("reads_for_overlaps")));
        }


        int n$ = 0;
        for (long i = 0; i < this.text.length(); i++) {
            if (this.text.get(i) == $index) {
                n$++;
            }
        }
        readsNumber = n$;    // assuming one $ per read

        readBegin = new long[n$ + 1];
        readBegin[0] = 0;
        int j = 1;
        for (long i = 0; i < this.text.length(); i++) {
            if (this.text.get(i) == $index) {
                readBegin[j] = i + 1;
                j++;
            }
        }
    }
    

    /**
     * Finds overlaps in all buckets. <br>
     * Loads suffix array buckets from files and saves overlaps to files. <br>
     *  
     * Attention! It doesn't try to fix errors in prefix!
     */
    public void findOverlaps() throws IOException {
        HashMap<Integer, ArrayList<Integer>> readsByBucket = divideReadsIntoBuckets();
        
        for (int bucket : readsByBucket.keySet()) {
            Timer t = new Timer();
            System.err.println("Loading suffix array bucket ...");
            SuffixArray sa = BucketsSorter.loadSuffixArrayBucket(text, new File(bucketsDir), bucket);
            System.err.println("Done in time = " + t);
            
            findOverlaps(bucket, readsByBucket.get(bucket), sa, null);
        }
    }
    
    /**
     * Finds overlaps in all buckets. <br>
     * Found overlaps are saved to overlaps class. <br>
     *  
     * Attention! It doesn't try to fix errors in prefix!
     */   
    public void findOverlaps(SuffixArray[] sas, Overlaps overlaps) throws IOException {
        HashMap<Integer, ArrayList<Integer>> readsByBucket = divideReadsIntoBuckets();
        
        for (int bucket : readsByBucket.keySet()) {
            findOverlaps(bucket, readsByBucket.get(bucket), sas[bucket], overlaps);
        }
    }
    
    /**
     * Finds overlaps in one bucket. <br>
     * if (overlaps != null) then found overlaps are saved to overlaps class,
     * else found overlaps are written to files.<br>
     *  
     * Attention! It doesn't try to fix errors in prefix!
     */
    void findOverlaps(int bucket, ArrayList<Integer> reads, SuffixArray sa, Overlaps overlaps) throws IOException {
//        System.out.println("Bucket = " + bucket + ", reads size = " + reads.size() + ", SA length = " + sa.length);

        Timer t = new Timer();
        System.out.println("Sorting reads ...");
        Collections.sort(reads, new ReadsComparator());
        System.out.println("Done in time = " + t);
        
        t.start();
        System.out.println("Searching overlaps ...");
        PatientExecutorService executor = new PatientExecutorService(availableProcessors);

        DedicatedWriter writer = null;
        if (overlaps == null) {
            writer = new DedicatedWriter(output + "." + bucket);
            new Thread(writer).start();
        }
        OverlapTaskContext overlapTaskContext = new OverlapTaskContext(
                text,
                reads,
                readBegin,
                readsNumber,
                sa,
                executor,
                writer,
                overlaps,
                numberOfErrors,
                errorsWindowSize,
                minOverlap
        );

        IntervalList readsInterval = new IntervalList();
        int[] lastErrors = new int[numberOfErrors];
        Arrays.fill(lastErrors, -errorsWindowSize);
        readsInterval.add(0, reads.size(), lastErrors);

        executor.execute(new OverlapTask(overlapTaskContext, 0, sa.length, readsInterval, 0, 5));

        try {
            executor.waitForShutdown();
        } catch (InterruptedException e) {
            e.printStackTrace(System.err);
        }
        if (writer != null) {
            writer.close();
        }
        System.out.println("Done in time = " + t);
    }
    
    HashMap<Integer, ArrayList<Integer>> divideReadsIntoBuckets() {
        HashMap<Integer, ArrayList<Integer>> readsByBucket = new HashMap<Integer, ArrayList<Integer>>();
        for (int read = 0; read < readsNumber; read++) {
            long rb = readBegin[read];
            int bucket = BucketsDivider.getBucketNumber(text, rb, bucketCharsNumber);
            if (!readsByBucket.containsKey(bucket)) {
                readsByBucket.put(bucket, new ArrayList<Integer>());
            }
            readsByBucket.get(bucket).add(read);
        }
        return readsByBucket;
    }

    private class ReadsComparator extends IntComparator {
        public int compare(int a, int b) {
            long aLen = readBegin[a + 1] - readBegin[a];
            long bLen = readBegin[b + 1] - readBegin[b];
            long aBegin = readBegin[a];
            long bBegin = readBegin[b];
            long len = Math.min(aLen, bLen);
            for (long i = 0; i < len; ++i) {
                int ca = text.get(aBegin + i);
                int cb = text.get(bBegin + i);
                if (ca != cb)
                    return ca - cb;
            }
            if (aLen == bLen)
                return 0;
            return (aLen < bLen) ? -1 : 1;
        }

    }

    public static boolean before(int read1, int read2, long shift) {
        if (shift != 0)
            return shift > 0;

        assert read1 != read2;
        return read1 < read2;
    }
    

}
