package ru.ifmo.genetics.tools.longReadsAssembler.suffixArray;

import static ru.ifmo.genetics.tools.longReadsAssembler.GluedDnasString.ALPHABET;

import java.io.File;
import java.io.IOException;
import java.util.Arrays;

import org.apache.commons.configuration.Configuration;

import ru.ifmo.genetics.statistics.Timer;
import ru.ifmo.genetics.tools.longReadsAssembler.GluedDnasString;
import ru.ifmo.genetics.tools.longReadsAssembler.arrays.FiveByteArrayWriter;

public class BucketsDivider implements Runnable {
    final int bucketCharsNumber;
    final int bucketsNumber;
    
    final String fullStringFileName;
    final String bucketsDir;
    
    public static void main(String[] args) throws IOException {
        Configuration config = BucketsSorter.getSABuilderConfiguration(args);
        if (config == null) {
            return;
        }
        
        BucketsDivider divider = new BucketsDivider(config);
        divider.run();
    }
    
    public BucketsDivider(Configuration config) throws IOException {
        bucketCharsNumber = config.getInt("bucket_chars_number");
        bucketsNumber = (int) Math.pow(ALPHABET, bucketCharsNumber);
        bucketsDir = config.getString("buckets_dir");
        fullStringFileName = config.getString("reads_for_overlaps");
    }
    
    @Override
    public void run() {
        try {
            GluedDnasString fullString = new GluedDnasString(new File(fullStringFileName));
            
            System.gc();
            
            System.out.println("Full string length = " + fullString.length());
            System.out.println("Free memory = " + Runtime.getRuntime().freeMemory() / (1 << 20) + " Mb");
    
            divideIntoBuckets(fullString, new File(bucketsDir));
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    /**
     * Divides all text suffixes into buckets and writes them to files bucketX.high/low in dir.
     */
    private void divideIntoBuckets(GluedDnasString text, File dir) throws IOException {
        Timer t = new Timer();
        
        if (!dir.exists()) {
            dir.mkdir();
        }
        for (File f : dir.listFiles()) {
            f.delete();
        }
        
        FiveByteArrayWriter[] writers = new FiveByteArrayWriter[bucketsNumber];
        for (int i = 0; i < bucketsNumber; i++) {
            writers[i] = new FiveByteArrayWriter(dir.toString() + "/bucket" + i);
        }
        
        int[] count = new int[bucketsNumber];
        for (long i = 0; i < text.length(); i++) {
            int bn = getBucketNumber(text, i);
            if (count[bn] == Integer.MAX_VALUE) {
                throw new RuntimeException("Size of bucket " + bn + " is larger than Integer.MAX_VALUE !");
            }
            count[bn]++;

            writers[bn].write(i);
        }
        
        for (int i = 0; i < bucketsNumber; i++) {
            writers[i].close();
        }
        
        System.err.println("Done, it took " + t);
    }
    
    /**
     * Divides all text suffixes into buckets.
     */
    public SuffixArray[] divideIntoBuckets(GluedDnasString text) {
        int[] count = new int[bucketsNumber];
        for (long i = 0; i < text.length(); i++) {
            int bucket = getBucketNumber(text, i);
            if (count[bucket] == Integer.MAX_VALUE) {
                throw new RuntimeException("Size of bucket " + bucket + " is larger than Integer.MAX_VALUE !");
            }
            count[bucket]++;
        }
        
        SuffixArray[] p = new SuffixArray[bucketsNumber];
        for (int i = 0; i < bucketsNumber; i++) {
            p[i] = new SuffixArray(text, count[i]);
        }
        
        Arrays.fill(count, 0);
        for (long i = 0; i < text.length(); i++) {
            int bucket = getBucketNumber(text, i);
            p[bucket].array.set(count[bucket], i);
            count[bucket]++;
        }
        
        return p;
    }
    
    
    public static int getBucketNumber(GluedDnasString text, long startPosInText, int charsNumber) {
        int res = 0;
        for (int i = 0; i < charsNumber; i++) {
            int v = text.getWithLastZeros(startPosInText + i);
            assert v != -1;
            res = res * ALPHABET + v;
        }
        return res;
    }

    static final byte $code = GluedDnasString.transform['$'];
    /**
     * Checks if bucket's chars specified by bucket number contain char '$'.
     */
    public static boolean bucketCharsContain$(int bucketN, int charsNumber) {
        for (int i = 0; i < charsNumber; i++) {
            int curChar = bucketN % ALPHABET;
            if (curChar == $code) {
                return true;
            }
            bucketN /= ALPHABET;
        }
        return false;
    }
    
    int getBucketNumber(GluedDnasString text, long pos) {
        return getBucketNumber(text, pos, bucketCharsNumber);
    }
}
