package ru.ifmo.genetics.tools.longReadsAssembler.suffixArray;

import static ru.ifmo.genetics.tools.longReadsAssembler.GluedDnasString.ALPHABET;
import static ru.ifmo.genetics.tools.longReadsAssembler.GluedDnasString.$index;

import java.io.*;
import java.util.ArrayList;
import java.util.Collections;
import java.util.concurrent.Executor;

import org.apache.commons.cli.*;
import org.apache.commons.configuration.*;

import ru.ifmo.genetics.statistics.Timer;
import ru.ifmo.genetics.tools.Util;
import ru.ifmo.genetics.tools.executors.PatientExecutorService;
import ru.ifmo.genetics.tools.longReadsAssembler.GluedDnasString;

public class BucketsSorter {
    final int availableProcessors;
    
    final int bucketCharsNumber;
    final int bucketsNumber;

    final static int smallBucketCharsNumber = 3;
    final static int smallBucketsNumber = (int) Math.pow(ALPHABET, smallBucketCharsNumber);
    
    final String fullStringFileName;
    final String bucketsDir;

    public static void main(String[] args) throws IOException, InterruptedException {
        Configuration config = getSABuilderConfiguration(args);
        if (config == null) {
            return;
        }
        
        BucketsSorter sorter = new BucketsSorter(config);
        sorter.run();
    }
    
    static Configuration getSABuilderConfiguration(String[] args) {
        Options options = new Options();
        options.addOption("h", "help", false, "prints this message");
        options.addOption("c", "config", true,  "sets the config file, default to config.properties");
        options.addOption("r", "reads", true, "sets the file with glued reads");
        options.addOption("b", "buckets-dir", true, "sets the buckets directory");
        options.addOption("B", "bucket-chars-number", true, "sets the prefix calculateSize for buckets");

        CommandLineParser parser = new PosixParser();
        CommandLine cmd;
        try {
            cmd = parser.parse(options, args);
        } catch (ParseException e) {
            e.printStackTrace(System.err);
            return null;
        }

        if (cmd.hasOption("help")) {
            new HelpFormatter().printHelp("overlap", options);
            return null;
        }

        String configFileName = cmd.getOptionValue("config", "config.properties");
        Configuration config;
        try {
            config = new PropertiesConfiguration(configFileName).subset("overlapper");
        } catch (ConfigurationException e) {
            e.printStackTrace(System.err);
            return null;
        }

        Util.addOptionToConfig(cmd, config, "reads", "reads_for_overlaps");
        Util.addOptionToConfig(cmd, config, "buckets-dir");
        Util.addOptionToConfig(cmd, config, "bucket-chars-number");
        
        return config;
    }

    public BucketsSorter(Configuration config) throws IOException {
        bucketCharsNumber = config.getInt("bucket_chars_number");
        bucketsNumber = (int) Math.pow(ALPHABET, bucketCharsNumber);
        availableProcessors = config.getInt("available_processors", Runtime.getRuntime().availableProcessors());
        bucketsDir = config.getString("buckets_dir");
        fullStringFileName = config.getString("reads_for_overlaps");
    }


    public void run() throws IOException, InterruptedException {
        GluedDnasString text = new GluedDnasString(new File(fullStringFileName));
        System.gc();

        System.err.println("Full string length = " + text.length());
        System.err.println("Free memory = " + Runtime.getRuntime().freeMemory() / (1 << 20) + " Mb");

        sortAllBuckets(text, new File(bucketsDir));
    }

    /**
     * Loads all suffix buckets, sorts and saves them to files sbucketX.high/low in dir.
     */
    private void sortAllBuckets(GluedDnasString text, File dir) throws IOException, InterruptedException {
        System.err.println("Sorting all buckets...");
        Timer t = new Timer();
        
        for (int i = 0; i < bucketsNumber; i++) {
            System.err.println("Bucket " + i + ", loading...");
            SuffixArray sa = new SuffixArray(text, dir.toString() + "/bucket" + i);
            System.err.println("SA bucket length = " + sa.length);

            System.err.println("Sorting...");
            Timer ct = new Timer();
            sort(sa);
            System.err.println("Done in time = " + ct);

            System.err.println("Saving...");
            sa.save(dir.toString() + "/sbucket" + i);
            sa = null;
        }
        
        System.err.println("Done, it took " + t);
    }

    /**
     * Sorts all suffix arrays.
     */
    public void sortAll(SuffixArray[] sas) throws InterruptedException {
        for (SuffixArray sa : sas) {
            sort(sa);
        }
    }
    
    /**
     * Sorts all suffixes in suffix array.
     */
    public void sort(SuffixArray sa) throws InterruptedException {
        PatientExecutorService executor = new PatientExecutorService(availableProcessors);
        bucketSort(executor, sa, 0, sa.length, 0);
        executor.waitForShutdown();
    }
    

    private void bucketSortWithForking(final Executor executor, final SuffixArray sa, final int begin,
                                       final int len, final int charsSorted) {
        if (len > 10000) {
            executor.execute(new Runnable() {
                @Override
                public void run() {
                    bucketSort(executor, sa, begin, len, charsSorted);
                }
            });
        } else {
            bucketSort(executor, sa, begin, len, charsSorted);
        }
    }

    /**
     * Sorts large arrays
     */
    private void bucketSort(final Executor executor, final SuffixArray sa, int begin, int len, int charsSorted) {
        // calculation
        int[] count = new int[smallBucketsNumber];
        for (int i = begin; i < begin + len; i++) {
            int b = BucketsDivider.getBucketNumber(sa.text, sa.get(i) + charsSorted, smallBucketCharsNumber);
            count[b]++;
        }

//        int maxC = 0;
//        for (int tc : count) {
//            maxC = Math.max(maxC, tc);
//        }
//        if (maxC > (len >> 1)) {
//            // buckets sort aren't effective now
//            quickSort(sa, begin, len, charsSorted);
//            return;
//        }

        int[] bi = new int[smallBucketsNumber];
        bi[0] = begin;
        for (int i = 1; i < smallBucketsNumber; i++) {
            bi[i] = bi[i - 1] + count[i - 1];
        }
        int[] cbi = bi.clone();

        int[] ei = new int[smallBucketsNumber];
        for (int i = 0; i < smallBucketsNumber; i++) {
            ei[i] = bi[i] + count[i];
        }

        // reordering
        int firstBN = 0;
        while (firstBN < smallBucketsNumber) {
            if (cbi[firstBN] == ei[firstBN]) {
                firstBN++;
                continue;
            }

            long cur = sa.get(cbi[firstBN]);
            int bn = BucketsDivider.getBucketNumber(sa.text, cur + charsSorted, smallBucketCharsNumber);
            while (bn != firstBN) {
                long ncur = sa.get(cbi[bn]);
                sa.set(cbi[bn], cur);
                cbi[bn]++;
                cur = ncur;
                bn = BucketsDivider.getBucketNumber(sa.text, cur + charsSorted, smallBucketCharsNumber);
            }
            sa.set(cbi[firstBN], cur);
            cbi[firstBN]++;
        }

        charsSorted += smallBucketCharsNumber;

        // sorting in small buckets
        for (int i = 0; i < smallBucketsNumber; i++) {
            if ((ei[i] - bi[i]) > 1) {
                boolean bucketCharsContain$ = BucketsDivider.bucketCharsContain$(i, smallBucketCharsNumber);
                if (!bucketCharsContain$) {
                    if ((ei[i] - bi[i]) < smallBucketsNumber) {
                        quickSort(sa, bi[i], ei[i] - bi[i], charsSorted);
                    } else {
                        bucketSortWithForking(executor, sa, bi[i], ei[i] - bi[i], charsSorted);
                    }
                }
            }
        }
    }

    private void quickSort(final SuffixArray sa, int begin, int len, int charsSorted) {
        // Insertion sort on smallest arrays
        if (len <= 7) {
            for (int i = begin; i < begin + len; i++)
                for (int j = i; j > begin && compare(sa.text, sa.get(j), sa.get(j - 1), charsSorted) < 0; j--)
                    sa.swap(j, j - 1);
            return;
        }

        // Choose a partition element, v
        int m = begin + (len >> 1); // middle element
        long v = sa.get(m);

        // Establish Invariant: v* (<v)* (>v)* v*
        int a = begin, b = a, c = begin + len - 1, d = c;
        int cr;
        while (true) {
            while (b <= c &&
                    ((cr = compare(sa.text, sa.get(b), v, charsSorted)) <= 0)) {
                if (cr == 0)
                    sa.swap(a++, b);
                b++;
            }
            while (c >= b &&
                    ((cr = compare(sa.text, sa.get(c), v, charsSorted)) >= 0)) {
                if (cr == 0)
                    sa.swap(c, d--);
                c--;
            }
            if (b > c)
                break;
            sa.swap(b++, c--);
        }

        int s, n = begin + len;
        s = Math.min(a - begin, b - a);
        sa.vecswap(begin, b - s, s);
        s = Math.min(d - c, n - d - 1);
        sa.vecswap(b, n - s, s);

        if ((s = b - a) > 1) {
            quickSort(sa, begin, s, charsSorted);
        }
        if ((s = d - c) > 1) {
            quickSort(sa, n - s, s, charsSorted);
        }
    }


    private int compare(GluedDnasString text, long pos1, long pos2, int charsSorted) {
        return fastCompareTill$(text, pos1 + charsSorted, pos2 + charsSorted);
    }

    /**
     * @param pos1 < length
     * @param pos2 < length
     */
    public static int fastCompareTill$(GluedDnasString array, long pos1, long pos2) {
        if (pos1 == pos2) {
            return 0;
        }

        while (true) {
            long b1 = array.getPackWithLastZeros(pos1);
            long b2 = array.getPackWithLastZeros(pos2);

            if (b1 != b2) {
                return (b1 < b2) ? -1 : 1;
            }
            if (GluedDnasString.nucsContain$(b1)) {
                return 0;
            }

            pos1 += GluedDnasString.nucsInLong;
            pos2 += GluedDnasString.nucsInLong;
        }

    }
    /**
     * @param pos1 < length
     * @param pos2 < length
     */
    public static int compareTill$(GluedDnasString array, long pos1, long pos2) {
        if (pos1 == pos2) {
            return 0;
        }

        while (true) {
            int b1 = array.get(pos1);
            int b2 = array.get(pos2);

            if (b1 != b2) {
                return b1 - b2;
            }
            if (b1 == $index) {
                return 0;
            }

            pos1++;
            pos2++;
        }
    }
    

    public static SuffixArray loadSuffixArrayBucket(GluedDnasString text, File bucketsDir, int bucket) throws IOException {
        File f = new File(bucketsDir, "sbucket" + bucket);
        SuffixArray sa = new SuffixArray(text, f.toString());
        return sa;
    }
    
}
