package ru.ifmo.genetics.tools.set;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.util.HashSet;
import java.util.Random;

public class BigLongsHashSet implements LongsHashSet {
    static final int smallCapacityPowerOf2 = 24;    // 16 M  [ * 8 byte = 0.13 Gb]
    static final long smallCapacity = 1 << smallCapacityPowerOf2;
    static final long smallCapacityMask = smallCapacity - 1;
    
    SmallLongsHashSet[] hs;
    long offset[];

    long capacity;
    long size = 0;
    
    static final int hashSizePowerOf2 = 16;
    static final int hashMask = (1 << hashSizePowerOf2) - 1;
    int[] m;

    public BigLongsHashSet() {
    }
    
    public BigLongsHashSet(long minCapacity, double loadFactor) {
        minCapacity = (long) (minCapacity / loadFactor);
        
        int cnt = (int) Math.ceil(minCapacity / (double) smallCapacity);
        capacity = smallCapacity * cnt;

        offset = new long[cnt];
        hs = new SmallLongsHashSet[cnt];
        for (int i = 0; i < cnt; i++) {
            hs[i] = new SmallLongsHashSet(smallCapacityPowerOf2);
            offset[i] = i * smallCapacity;
        }
        
        m = new int[1 << hashSizePowerOf2];
        for (int i = 0; i < m.length; i++) {
            m[i] = i % hs.length;
        }
    }
    
    /**
     * @param memoryUsage in bytes
     */
    public BigLongsHashSet(long memoryUsage) {
        this(memoryUsage / 8, 1);
    }
    
    int getIndex(long v) {
        int i = (int) (v ^ (v >>> 33));
        i = i ^ (i >>> 16);
        i = i & hashMask;
        return m[i];
    }
    
    @Override
    public boolean contains(long v) {
        int i = getIndex(v);
        return hs[i].contains(v);
    }

    @Override
    public boolean put(long v) {
        int i = getIndex(v);
        boolean res = hs[i].put(v);
        if (res) {
            size++;
        }
        return res;
    }
    
    @Override
    public long size() {
        return size;
    }

    @Override
    public long capacity() {
        return capacity;
    }

    @Override
    public void reset() {
        for (SmallLongsHashSet set: hs) {
            set.reset();
        }
        size = 0;
    }

    @Override
    public long getPosition(long v) {
        int i = getIndex(v);
        long p = hs[i].getPosition(v);
        return (p == -1) ? -1 : (p + offset[i]);
    }

    @Override
    public long elementAt(long i) {
        return hs[(int)(i >>> smallCapacityPowerOf2)].elementAt(i & smallCapacityMask);
    }
    
    // Test
    public static void main(String[] args) {
        int cnt = 1000000;
        BigLongsHashSet b = new BigLongsHashSet(500000000);
        Random r = new Random(832449587);
        
        System.out.println("Testing...");
        boolean g = true;
        
        HashSet<Long> els = new HashSet<Long>();
        for (int i = 0; i < cnt; i++) {
            long v = -1;
            while ((v == -1) || (els.contains(v))) {
                v = r.nextLong();
            }
            els.add(v);
        }
        
        for (long v : els) {
            g &= b.put(v);
        }
        
        g &= (b.size == els.size());
        
        for (long v : els) {
            g &= !b.put(v);
        }

        for (long v : els) {
            g &= b.contains(v);
        }

        for (int i = 0; i < cnt; i++) {
            long v = -1;
            while ((v == -1) || (els.contains(v))) {
                v = r.nextLong();
            }
            g &= !b.contains(v);
        }
        
        for (long v : els) {
            long p = b.getPosition(v);
            long gv = b.elementAt(p);
            g &= (gv == v);
        }
        
        for (int i = 0; i < cnt; i++) {
            long v = -1;
            while ((v == -1) || (els.contains(v))) {
                v = r.nextLong();
            }
            long p = b.getPosition(v);
            g &= (p == -1);
        }
        
        
        if (!g) {
            throw new AssertionError();
        }
        System.out.println("Test passed.");
    }

    @Override
    public void write(DataOutput out) throws IOException {
        out.writeInt(hs.length);
        for (int i = 0; i < hs.length; ++i) {
            hs[i].write(out);
        }
        out.writeInt(offset.length);
        for (int i = 0; i < offset.length; ++i) {
            out.writeLong(offset[i]);
        }

        out.writeLong(capacity);
        out.writeLong(size);
        out.writeInt(m.length);
        for (int i = 0; i < m.length; ++i) {
            out.writeInt(m[i]);
        }
    }

    @Override
    public void readFields(DataInput in) throws IOException {
        int hsl = in.readInt();
        hs = new SmallLongsHashSet[hsl];
        for (int i = 0; i < hs.length; ++i) {
            hs[i] = new SmallLongsHashSet(smallCapacityPowerOf2);
            hs[i].readFields(in);
            System.err.println(i + " out of " + hs.length + " done");
        }
        
        int ol = in.readInt();
        offset = new long[ol];
        for (int i = 0; i < offset.length; ++i) {
            offset[i] = in.readLong();
        }

        capacity = in.readLong();
        size = in.readLong();
        
        int ml = in.readInt();
        m = new int[ml];
        for (int i = 0; i < m.length; ++i) {
            m[i] = in.readInt();
        }
    }
}
